# ****************************************************
# Name des Moduls: bgui
# Name des Projekts: Bufferlo
#
# Autor des Moduls:
#       Thorsten Reichelt, Thorsten.Reichelt@web.de
#
# Aufgaben des Moduls: siehe DocString
# Compiler fr eingebettete Dokumentation: PyDoc
#
# Datum der letzten nderung: 16.01.2003
# ****************************************************

"""Dieses Modul ist das GUI des Bufferlo-Projekts. Das GUI nimmt
   Benutzeranfragen auf verschiedenen Wegen (Objekt anklicken, Prp-Phrase
   eingeben, Objektliste durchgehen) an, lsst das bufferlo-Modul die
   Berechnungen ausfhren und stellt dann die Ergebnisse auf der Karte da.

   Beim Aufruf des Moduls knnen die Parameter -i<inifile> oder -d bergeben
   werden:
   -i        ldt eine alternative Einstellungsdatei
   inifile   Datei mit BGui-Einstellungen
   -d        ldt BGui ohne Benutzereinstellungen

   Das Modul beinhalted 4 Klassen:
   Gui: Beinhaltet alles, was sich im Hauptfenster des GUI abspielt. Also alle
        Widgets, Menus, Kommandos, Binds und Darstellungsfunktionen. Lediglich
        die Fehler-, die Copyright- und die Hilfe-Anzeige sind ausgegliedert.
   BGuiError: Zeigt und zhlt Fehlermeldungen im eigenen Fenster.
   CopyRight: Zeigt die Copyright-Informationen im eigenen Fenster.
   Help: Zeigt die Farblegende und die Tastaturbelegung im eigenen Fenster.
"""            
   

from tkMessageBox import askokcancel
from tkFileDialog import askopenfilename
from bbox import *
from geomap import GeoMap
from bufferlo import *
from geoobject import ViewerObject
from berror import *
from Tkinter import *
import btools
import ptag
import Pmw
import ConfigParser
import os
import prepcontainer

#*************************************************************** Konstanten
NEW = "new"
REDRAW = "redraw"
ON = 1
OFF = 0
DEBUG = 1
XML = "XML"


class Gui:
    """ Beinhaltet alles, was sich im Hauptfenster des GUI abspielt. Also alle
        Widgets, Menus, Kommandos, Binds und Darstellungsfunktionen. Lediglich
        die Fehler-, die Copyright- und die Hilfe-Anzeige sind ausgegliedert.
        
        Zum Start des GUI langt es, eine Objektinstanz von Gui anzulegen.
        Dabei mssen folgende Parameter angeben werden:
        userini: Legt fest, aus welcher Datei (zustzlich zur defaults.ini)
                 Einstellungen geladen werden. Darf auch ungltige Dateinamen,
                 Leerstring oder None enthalten, dann werden einfach keine zus.
                 Einstellungen geladen.
    """
    
    #******************************************************** Initiierungen
    def __init__(self, userini):
        #********************************* __init__.1. Variablen initiieren
        self.userini = userini
        self.current_prep = None
        self.error_window = None
        self.showing_buffer = 0
        self.inactive_bfill = "#08F"

        #Ini-Datei lesen
        self.readIni()

        #****************************** __init__.2. Aufbau des GUI (Aufruf)
        self.root = Tk()
        Pmw.initialise(self.root)
        self.root.title("Bufferlo")

        self.makeWidgets()
        self.makeMenu()
        self.makeBindings()

        #************************** __init__.3. GUI-Startparameter anwenden
        #Jetzt eventuell Karte laden und Viewer setzen
        if self.start_map:
            self.loadMap(self.objdb, self.objtyp)
        else:
            self.karte=""

        if self.viewer:
            self.modifyViewer(state = ON)
        else:
            self.modifyViewer(state = OFF,
                              coords = self.inactive_viewer.getXY(),
                              dir = self.inactive_viewer.direction
                              )

        #****************************************** __init__.4. GUI starten
        self.root.mainloop()
        
    
    def readIni(self):
        """ Initialisiert eine Reihe von allgemeinen Variablen.
            Dazu werden die Dateien 'default.ini' (obligatorisch) und die in
            self.userini angegebene (optional) gelesen.

            DATENQUELLEN
            Momentan funktionieren nur xml-Dateien als Datenquellen.
            Die einzelnen Vars hier sind:
            self.prepdb  : Praepositions-Datenquelle
            self.preptyp : Format der prepdb
            self.objdb   : Kartenobjekt-Datenquelle
            self.objtyp  : Format der objdb
            self.gramdb  : Grammatik-Datenquelle
            self.gramtyp : Format der Grammatikdatenquelle

            ANZEIGEPARAMETER
            self.bfill     : Bufferfllfarbe
            self.ofill     : Objektfllfarbe
            self.bb_line   : Linienfarbe der sichtweisenabh. BB
            self.dbb_line  : Linienfarbe der sichtweisenunabh. BB
            self.start_map : Gleich zu Beginn eine Karte laden?
            self.zoom      : Aktueller Zoomfaktor
            self.zoomlist  : Liste der Zoomfaktore im DropDown-Menu

            VIEWER
            self.viewer_activate: Betrachter bei Programmstart aktivieren?
            self.set_viewer_dir : Betrachter mit Richtung oder ohne?
            self.viewer_name    : Name des Betrachters
            self.viewer_id      : ID des Betrachters
            self.viewer_coord   : Koordinaten d.B.
            self.viewer_dir     : Richtungsvektor des Betrachters
                                  (Achtung, wird auch gebraucht, wenn
                                  set_viewer_dir aus ist!)          
        """

        #***************************************** readIni.1. Dateien lesen
        config = ConfigParser.ConfigParser()
        config.readfp(open('defaults.ini'))
        config.read(self.userini)
        #Momentan mssen ALLE bentigten Werte in mindestens einer der
        #beiden Dateien vorkommen, da es sonst zu einer exception kommt.
        #Dies koennte mit einer der folgenden Methoden verhindert werden:
        #1. Vor jeder Wertzuweisung durch die get-Methoden ueberpruefen, ob
        #   die entsprechende Option gesetzt ist.
        #2. Mittels der Methoden add_section und set erstmal alle
        #   benoetigten Werte im config-Objekt setzen.
        #3. try-except-Klammern um jede einzelne Zuweisung, im except-Teil
        #   dann halt programminterne default-Zuweisungen.


        #********************************* readIni.2. Datenquellenparameter
        self.prepdb  = config.get("data sources", "preposition data")
        self.preptyp = config.get("data sources", "preposition data type")
        self.objdb   = config.get("data sources", "map objects data")
        self.objtyp  = config.get("data sources", "map objects type")
        self.gramdb  = config.get("data sources", "grammar data")        
        self.gramtyp = config.get("data sources", "grammar data type")
        
        #************************************** readIni.3. Anzeigeparameter
        self.bfill     = config.get(       "display", "Buffer Fill")
        self.ofill     = config.get(       "display", "Object Fill")
        self.bb_line   = config.get(       "display", "Bounding Box Line")
        self.dbb_line  = config.get(       "display", "Default Bounding Box Line")
        self.start_map = config.getboolean("display", "Load map on start")
        self.zoom      = config.getfloat(  "display", "zoom")
        self.zoomlist  = config.get(       "display", "zoomlist")

        self.bfill = self.bfill.strip('"')
        self.ofill = self.ofill.strip('"')
        self.bb_line = self.bb_line.strip('"')
        self.dbb_line = self.dbb_line.strip('"')
        self.zoomlist = self.zoomlist.replace(",", " ")
        self.zoomlist = self.zoomlist.replace(";", " ")
        self.zoomlist = self.zoomlist.split()
        for zaehler in range(len(self.zoomlist)):
            self.zoomlist[zaehler] = str(float(self.zoomlist[zaehler]))

        #*********************************** readIni.4. Betrachterparameter
        self.set_viewer_dir  = config.getboolean("viewer", "Set viewer dir")
        viewer_name          = config.get(       "viewer", "Viewer name")
        viewer_id            = config.get(       "viewer", "Viewer id")
        viewer_coords        = config.get(       "viewer", "Viewer coordinates")
        viewer_dir           = config.get(       "viewer", "Viewer direction")

        viewer_coords = viewer_coords.split("/")
        viewer_dir = viewer_dir.split("/")

        #Und jetzt noch self.viewer und self.inactive_viewer setzen
        if config.getboolean("viewer", "Activate Viewer"):
            self.inactive_viewer = None
            if self.set_viewer_dir:
                self.viewer = ViewerObject(viewer_name, viewer_id,
                                           [float(viewer_coords[0]), 
                                            float(viewer_coords[1])],
                                           [float(viewer_dir[0]),
                                            float(viewer_dir[1])]
                                           )
            else:
                self.viewer = ViewerObject(viewer_name, viewer_id,
                                           [float(viewer_coords[0]), 
                                            float(viewer_coords[1])]
                                           )
        else:
            self.viewer = None
            if self.set_viewer_dir:
                self.inactive_viewer = ViewerObject(viewer_name, viewer_id,
                                                    [float(viewer_coords[0]), 
                                                     float(viewer_coords[1])],
                                                    [float(viewer_dir[0]), 
                                                     float(viewer_dir[1])]
                                                    )
            else:
                self.inactive_viewer = ViewerObject(viewer_name, viewer_id,
                                                    [float(viewer_coords[0]), 
                                                     float(viewer_coords[1])]
                                                    )
                

    #****************************************** Aufbau des GUI (Funktionen)
                
    def makeMenu(self):
        """Erzeugt die Menleiste des GUI.
        self.root muss gesetzt sein."""
        
        topmenu = Menu(self.root)
        self.root.config(menu = topmenu)

        #************************************************ makeMenu.1. Datei
        mdatei = Menu(topmenu, tearoff = 0)
        mdatei.add_command(label = "Neue Karte laden...",
                           command = self.loadMapDialogue,  underline = 0)
        mdatei.add_command(label = "Prps laden...",
                           command = self.loadPrepDialogue, underline = 0)
        mdatei.add_command(label = "Grammatikdatei laden ...",
                           command = self.loadGramDialogue, underline = 0)
        mdatei.add_command(label = "Exit",
                           command = (lambda: self.closeWindow(self.root)),
                           underline = 0)
        topmenu.add_cascade(label = "Datei", menu = mdatei, underline = 0)

        #********************************************** makeMenu.2. Ansicht
        mansicht = Menu(topmenu, tearoff = 0)
        mansicht.add_command(label = "Refresh",
                             command = (lambda: self.updateMap(mode=REDRAW)),
                             underline = 0)
        topmenu.add_cascade(label="Ansicht", menu = mansicht, underline = 0)

        #************************************ makeMenu.3. Zoom (in Ansicht)
        mzoom = Menu(mansicht, tearoff = 0)
        mzoom.add_command(label = "25%",
                          command = (lambda: self.changeZoom(25.0)),
                          underline = 0)
        mzoom.add_command(label = "50%",
                          command = (lambda: self.changeZoom(50.0)),
                          underline = 0)
        mzoom.add_command(label = "75%",
                          command = (lambda: self.changeZoom(75.0)),
                          underline = 0)
        mzoom.add_command(label = "100%",
                          command = (lambda: self.changeZoom(100.0)),
                          underline = 0)
        mzoom.add_command(label = "125%",
                          command = (lambda: self.changeZoom(125.0)),
                          underline = 0)
        mzoom.add_command(label = "150%",
                          command = (lambda: self.changeZoom(150.0)),
                          underline = 0)
        mzoom.add_command(label = "200%",
                          command = (lambda: self.changeZoom(200.0)),
                          underline = 0)        
        mansicht.add_cascade(label = "Zoom", menu = mzoom, underline = 0)

        #************************************************ makeMenu.4. Hilfe
        mhilfe = Menu(topmenu, tearoff = 0)
        mhilfe.add_command(label = "Hilfe",
                           command = self.showHelp,      underline = 0)
        mhilfe.add_command(label = "Copyright",
                           command = self.showCopyRight, underline = 0)
        topmenu.add_cascade(label = "Hilfe", menu=mhilfe, underline = 0)


    def makeWidgets(self):
        """ Erzeugt smtliche widgets des GUI.
            self.root muss gesetzt sein. """
        
        #****************** makeWidgets.1. Grobeinteilung in vier Einheiten
        #1. Kartenanzeigefeld        
        self.canvasframe = Frame(self.root)
        self.canvasframe.pack(side=TOP, expand=YES, fill=BOTH)

        #2. Informations- und Bedienzone        
        dframe_left = Frame(self.root, bd=2, relief=RIDGE, bg="grey")
        dframe_left.pack(side=LEFT, expand=YES, fill=BOTH, pady=5)
        dframe_middle = Frame(self.root, bd=2, bg="black", relief=RIDGE)
        dframe_middle.pack(side=LEFT, expand=YES, fill=BOTH, padx=6, pady=5)
        dframe_right = Frame(self.root, bd=2, relief=RIDGE, bg="grey")
        dframe_right.pack(side=LEFT, expand=YES, fill=BOTH, pady=5)

        #Kartenanzeigefeld, wird ganz unten packed
        self.canvas = Pmw.ScrolledCanvas(self.canvasframe,
                                         usehullsize = 1,
                                         hull_width = 600,
                                         hull_height = 400,
                                         canvasmargin = 50)
        self.canvas.interior().configure(bg ="#FFFFF8")

        #****************************** makewidgets.2. Linkes Dialogfenster
        #**************************************** makewidgets.2.1. Eingabe
        entryframe = Frame(dframe_left,bd=2,relief=RIDGE)
        entryframe.pack(expand=YES, fill=BOTH)
        
        self.prep_entry = Entry(entryframe)
        self.prep_entry.insert(INSERT, "Bitte Prpositionalphrase eingeben")
        self.prep_entry.focus()
        self.prep_entry.selection_range(0,END)
        self.prep_entry.pack(expand=YES, fill=X)
        
        prep_button_show = Button(entryframe, text="Zeige Buffer", command = self.parsePrepEntry)
        prep_button_show.pack(side=LEFT)
        
        prep_button_clear = Button(entryframe, text="Lsche Eintrag", command = self.clearPrepEntry)
        prep_button_clear.pack(side=RIGHT)

        #****************************************** makeWidgets.2.2. Viewer
        viewer_group = Pmw.Group(dframe_left, tag_text="Betrachter")
        viewer_group.pack(expand=YES, fill=BOTH)

        viewer_setting_frame = Frame(viewer_group.interior())
        viewer_setting_frame.pack(side=BOTTOM, expand=YES, fill=BOTH)

        viewer_button_apply = Button(viewer_setting_frame,
                                     text = "Anwenden",
                                     command = (lambda: self.modifyViewer(state=ON))
                                     )
        viewer_button_apply.pack(side=RIGHT, expand=YES, padx=5)

        viewer_coord_group = Pmw.Group(viewer_setting_frame,
                                       tag_text = "Koordinaten"
                                       )
        viewer_coord_group.pack(side=TOP, expand=YES, fill=BOTH)
        self.viewer_ycoord_entry = Pmw.EntryField(viewer_coord_group.interior(),
                                                  labelpos = "w",
                                                  label_text = "Y:",
                                                  validate = {"validator" : "real"},
                                                  command = (lambda: self.modifyViewer(state=ON))
                                                  )
        self.viewer_ycoord_entry.component("entry").config(width=10)
        self.viewer_ycoord_entry.pack(side=RIGHT, expand=YES, fill=BOTH, pady=2, padx=2)
        self.viewer_xcoord_entry = Pmw.EntryField(viewer_coord_group.interior(),
                                                  labelpos = "w",
                                                  label_text = "X:",
                                                  validate = {"validator" : "real"},
                                                  command = (lambda: self.viewer_ycoord_entry.component("entry").focus())
                                                  )
        self.viewer_xcoord_entry.component("entry").config(width=10)
        self.viewer_xcoord_entry.pack(side=RIGHT, expand=YES, fill=BOTH, pady=2, padx=2)

        #Falls in Ini gewscht, Bedienfeld fr Betrachterrichtung aufbauen
        if self.set_viewer_dir:
            viewer_dir_group = Pmw.Group(viewer_setting_frame,
                                         tag_text = "Richtungsvektor"
                                         )
            viewer_dir_group.pack(side=TOP, expand=YES, fill=BOTH)
            self.viewer_xdir_entry = Pmw.EntryField(viewer_dir_group.interior(),
                                                    labelpos = "w",
                                                    label_text = "X:",
                                                    validate = {"validator" : "real"},
                                                    )
            self.viewer_xdir_entry.component("entry").config(width=10)
            self.viewer_xdir_entry.pack(side=LEFT, expand=YES, fill=BOTH, pady=2, padx=2)
            self.viewer_ydir_entry = Pmw.EntryField(viewer_dir_group.interior(),
                                                    labelpos = "w",
                                                    label_text = "Y:",
                                                    validate = {"validator" : "real"},
                                                    command = (lambda: self.modifyViewer(state=ON))
                                                    )
            self.viewer_ydir_entry.component("entry").config(width=10)
            self.viewer_ydir_entry.pack(side=LEFT, expand=YES, fill=BOTH, pady=2, padx=2)

        #Ueber die Eingabefelder kommen jetzt noch aktiv, Zentrieren und Setzen
        self.viewer_state = IntVar()
        viewer_checkbox_active = Checkbutton(viewer_group.interior(),
                                             text="aktiv",
                                             var=self.viewer_state,
                                             command = (lambda: self.modifyViewer())
                                             )
        viewer_checkbox_active.pack(side=LEFT, expand=YES)

        viewer_button_center = Button(viewer_group.interior(),
                                      text="Zentrieren",
                                      command = (lambda: self.centerOnXY(self.viewer.x,self.viewer.y))
                                      )
        viewer_button_center.pack(side=LEFT, expand=YES)

        viewer_button_set = Button(viewer_group.interior(),
                                   text="Setze",
                                   command = (lambda: self.setViewer())
                                   )
        viewer_button_set.pack(side=RIGHT, expand=YES)       


        #*************************** makeWidgets.3. Mittleres Dialogfenster
        self.co_display = Label(dframe_middle, text="X:   Y:  ", bg="black", fg="red", width=5)
        self.co_display.pack(expand=YES, fill=BOTH)

        self.messages = Pmw.MessageBar(dframe_middle,
                                       labelpos = E,
                                       label_text = "Status")
        self.messages.component("entry").config(bg = "black", fg = "yellow")
        self.messages.component("label").config(bg = "black", fg = "red")
        self.messages.component("hull").config(bg = "black")
        self.messages.message("userevent", "Welcome to Bufferlo")
        self.messages.pack(expand=YES, fill=BOTH)

        self.binfo_prep = Pmw.ScrolledField(dframe_middle,
                                            labelpos = E,
                                            label_text = "Prposition",
                                            text = "Keine Prp")
        self.binfo_prep.component("entry").config(bg = "black", fg = "yellow", width=12)
        self.binfo_prep.component("label").config(bg = "black", fg = "red")
        self.binfo_prep.component("hull").config(bg = "black")
        self.binfo_prep.pack(expand=YES, fill=Y, anchor=W)
        self.binfo_obj = Pmw.ScrolledField(dframe_middle,
                                            labelpos = E,
                                            label_text = "Objekt",
                                            text = "Kein Objekt")
        self.binfo_obj.component("entry").config(bg = "black", fg = "yellow", width=12)
        self.binfo_obj.component("label").config(bg = "black", fg = "red")
        self.binfo_obj.component("hull").config(bg = "black")
        self.binfo_obj.pack(expand=YES, fill=Y, anchor=W)
       
        self.mapinfo = Message(dframe_middle, text="Noch keine Karte geladen!\n\n\n", aspect=300)
        self.mapinfo.config(bg="black",fg="lightgrey")
        self.mapinfo.pack(expand=YES, fill=BOTH)

        #***************************** makeWidgets.4. Rechtes Dialogfenster
        zoom_group = Pmw.Group(dframe_right, tag_text="Zoom")
        zoom_group.pack(expand=YES, fill=BOTH)
        misc_group = Pmw.Group(dframe_right, tag_text="Bedienung")
        misc_group.pack(expand=YES, fill=BOTH)

        #Die Zoomauswahl erstellen
        self.zoom_entry = Pmw.ComboBox(zoom_group.interior(),
                                       labelpos = "w",
                                       label_text = "Zoom (%):",
                                       scrolledlist_items = self.zoomlist,
                                       selectioncommand = self.changeZoom
                                       )
        self.zoom_entry.component("entryfield").configure(validate = {'validator' : 'real'})
        #Ini-Zoom anwenden, und gegebenenfalls in die Liste einfgen
        if not (str(self.zoom*100) in self.zoomlist):
            self.zoom_entry.insert(END, str(self.zoom*100))
            #Sortiertes Einfgen wre schn, aber da die Liste ein Tupel ist,
            #kann man die Reihenfolge nicht direkt ndern

        self.zoom_entry.selectitem(str(self.zoom*100))
        self.zoom_entry.pack()

        self.button1_choice = Pmw.RadioSelect(misc_group.interior(),
                                              buttontype = "radiobutton",
                                              labelpos = "w",
                                              label_text = "Linksklick macht:",                                              
                                              orient = "vertical",
                                              command = (lambda name: self.changeButton1(name))
                                              )
        self.button1_choice.pack()
        self.button1_choice.add("cycleobject",
                                text = "nchstes Objekt"
                                )
        self.button1_choice.add("chooseobject",
                                text = "whle Objekt"
                                )
        self.button1_choice.invoke("cycleobject")

        self.canvas.pack(expand=YES, fill=BOTH)

        self.balloon = Pmw.Balloon(self.root,
                                   relmouse = BOTH,
                                   statuscommand = self.messages.helpmessage
                                   )

    #************************************************************* Bindings
    def changeButton1(self, name):
        """ndert das Button1-Binding je nach dem Wert von <name>, so dass
           ein Linksklick entweder ein Objekt weiter geht, oder ein Objekt
           auf der Karte direkt auswhlt.
           Wird vom Links-Klick-Radiobutton aufgerufen."""

        if name == "cycleobject":
            self.canvas.interior().bind('<Button-1>',   lambda event: self.cycleObjects(1))
        elif name == "chooseobject":
            self.canvas.interior().bind('<Button-1>',   self.chooseObject)
            
        
    def makeBindings(self):
        """Die GUI-Bindings (Tastatur- und Maus-Kommandozuweisungen.)"""
        
        self.root.bind("<F1>", lambda event: self.hotKeys(event))
        self.root.bind("<F2>", lambda event: self.hotKeys(event))
        self.root.bind("<F3>", lambda event: self.hotKeys(event))
        self.root.bind("<F4>", lambda event: self.hotKeys(event))
        self.root.bind("<F5>", lambda event: self.hotKeys(event))
        self.root.bind("<F6>", lambda event: self.hotKeys(event))
        self.root.bind("<F7>", lambda event: self.hotKeys(event))
        self.root.bind("<F8>", lambda event: self.hotKeys(event))
        
        self.canvas.interior().bind("<Motion>",     self.showCoordinates)
        self.canvas.interior().bind("<Button-3>",   lambda event: self.cycleObjects(-1))
        
        self.prep_entry.bind('<Return>', lambda event: self.parsePrepEntry())


    #******************************************************** Menkommandos
    def closeWindow(self, win):
        """Schliet das bergebene Fenster"""
        if askokcancel('Beenden besttigen', "Bufferlo wirklich verlassen?"): win.destroy()

    def loadGramDialogue(self):
        """ffnet den Standard-Datei-Laden Dialog zum Aussuchen einer Grammatikdatei"""

        filename = askopenfilename(filetypes=[("XML-Dateien",".xml"), ("Alle Dateien","*")])
        if filename:
            self.loadGram(filename, XML)

    def loadMapDialogue(self):
        """ffnet den Standard-Datei-Laden Dialog zum Aussuchen einer Karte"""

        filename = askopenfilename(filetypes=[("XML-Dateien",".xml"), ("Alle Dateien","*")])
        if filename:
            self.loadMap(filename, XML)

    def loadPrepDialogue(self):
        """ffnet den Standard-Datei-Laden Dialog zum Aussuchen einer
           Prpositionsdatei."""

        filename = askopenfilename(filetypes=[("XML-Dateien",".xml"), ("Alle Dateien","*")])
        if filename:
            self.loadPreps(filename, XML)
        
    #*************************************************************** Events
    def chooseObject(self, event):
        """Findet heraus, auf welches Objekt auf der Karte der Benutzer
           geklickt hat, macht es zum aktuellen Objekt und veranlasst das
           Zeichnen des dazugehrigen Buffers."""
        
        #Erstmal die Canvaskoordinaten in echte Koordinaten umwandeln
        x = self.canvas.canvasx(event.x)
        y = self.canvas.canvasy(event.y)

        #Informationen ber das angeklickte Object sammeln
        #(find_closest findet auch dann ein Objekt, wenn nur in die
        #Nhe geklickt wurde)
        objectID = self.canvas.find_closest(x, y)
        tags = self.canvas.gettags(objectID)

        #Der erste tag eines Kartenobjektes beginnt immer mit "ID"
        if tags[0][0:2] == "ID":
            #Eventuelle Fllung des bisher aktuellen Objekts lschen
            self.canvas.interior().itemconfigure("ID"+self.karte.cobjects[self.current_object].id,
                                                 fill = "")
            #Neues Objekt in der Liste der geladenen Objekte wiederfinden
            self.current_object = 0
            max = len(self.karte.cobjects)
            while (self.karte.cobjects[self.current_object].id != tags[0][2:] ) and (self.current_object < max):
                self.current_object += 1
            
            #Infofeld aktualisieren und neues Objekt eventuell fllen
            self.binfo_obj.configure(text = self.karte.cobjects[self.current_object].name)
            self.canvas.interior().itemconfigure("ID"+self.karte.cobjects[self.current_object].id,
                                                 fill = self.ofill)

            self.drawBuffer(self.karte.cobjects[self.current_object])
                

    def hotKeys(self, event):
        """F-Tasten Verwaltung. Steuert die Behandlung der Tasten
           F1-F8."""

        if event.keysym == "F1":
            #Hilfetext anzeigen
            help = Help()
        elif event.keysym == "F2":
            #Buffer fllen an/aus
            if self.bfill:
                self.inactive_bfill = self.bfill
                self.bfill = ""
            else:
                self.bfill = self.inactive_bfill
                self.inactive_bfill = ""
            if self.showing_buffer:
                self.drawBuffer(self.karte.cobjects[self.current_object])
        elif event.keysym == "F3":
            #vorheriges Objekt
            self.cycleObjects(-1)
        elif event.keysym == "F4":
            #nchstes Objekt
            self.cycleObjects(1)
        elif event.keysym == "F5":
            #Karte aktualisieren
            self.updateMap(mode = REDRAW)
        elif event.keysym == "F6":
            #sichtweisenabhngige BB an/aus
            objectID = self.canvas.interior().find_withtag("bbox")
            if self.canvas.interior().itemcget(objectID, "outline"):
                self.canvas.interior().itemconfigure(objectID, outline = "")
                objectIDs = self.canvas.interior().find_withtag("bbox-text")
                for id in objectIDs:
                    self.canvas.interior().itemconfigure(id, fill = "")
            else:
                self.canvas.interior().itemconfigure(objectID, outline = self.bb_line)
                objectIDs = self.canvas.interior().find_withtag("bbox-text")
                for id in objectIDs:
                    self.canvas.interior().itemconfigure(id, fill = "black")
        elif event.keysym == "F7":
            #sichtweisenunabhngige BB an/aus
            objectID = self.canvas.interior().find_withtag("dbbox")
            if self.canvas.interior().itemcget(objectID, "outline"):
                self.canvas.interior().itemconfigure(objectID, outline = "")
                objectIDs = self.canvas.interior().find_withtag("dbbox-text")
                for id in objectIDs:
                    self.canvas.interior().itemconfigure(id, fill = "")
            else:
                self.canvas.interior().itemconfigure(objectID, outline = self.dbb_line)
                objectIDs = self.canvas.interior().find_withtag("dbbox-text")
                for id in objectIDs:
                    self.canvas.interior().itemconfigure(id, fill = "black")
        elif event.keysym == "F8":
            #Aktuelles Objekt fllen an/aus
            objectID = self.canvas.interior().find_withtag("ID" + self.karte.cobjects[self.current_object].id)
            if self.canvas.interior().itemcget(objectID, "fill"):
                self.canvas.interior().itemconfigure(objectID, fill = "")
            else:
                if not self.ofill: self.ofill = "red"
                self.canvas.interior().itemconfigure(objectID, fill = self.ofill)
                
            
    def showCoordinates(self, event):
        """Zeigt die echten Koordinaten an, an denen der Mauszeiger steht."""
        x = self.canvas.canvasx(event.x) / self.zoom
        y = self.canvas.canvasy(event.y) / self.zoom
        x, y = round(x,1), round(y,1)
        string = "X: " + str(x) + "  Y: " + str(y)
        self.co_display.config(text = string)


    #--------------------------------------------callback-handlers (ohne events)
    def changeZoom(self, zoom):
        """ndert den Zoomfaktor und wendet ihn auf die Karte an.
           <zoom> wird dabei als Prozentwert erwartet (und msste hier
           theoretisch immer als String in der Form 'xx.x' ankommen)."""

        self.zoom = float(zoom)/100

        #Falls ein neuer Zoomfaktor eingegeben wurde, wollen wir ihn in die
        #Zoomauswahlliste bernehmen
        if not (str(zoom) in self.zoom_entry.getvalue()):
            if not (str(zoom) in self.zoom_entry.component("scrolledlist").get()):
                self.zoom_entry.insert(END, str(self.zoom*100))
            self.zoom_entry.selectitem(str(float(zoom)))

        #Jetzt noch Karte und Viewer neu zeichnen
        self.updateMap(mode=REDRAW)
        self.modifyViewer()

        
    def clearPrepEntry(self):
        """Lscht alle Zeichen in <self.prep_entry> und alle Buffer und
           BBs auf dem Canvas."""
        
        self.prep_entry.delete(0,END)
        self.canvas.delete("bbox_all")
        self.canvas.delete("buffer")
        self.showing_buffer = 0
        

    def parsePrepEntry(self):
        """Liest Input aus <self.prep_entry>, ruft damit dann die Funktion
           <self.prepParser> auf, whlt anhand des Ergebnisses die neue
           Prposition und das neue Objekt aus und veranlasst die
           Darstellung.
        """

        #Den Eintrag im Praepositionsfeld lesen und auswerten
        prep_entry = self.prep_entry.get()
        prep_entry.strip()

        if prep_entry:
            self.messages.message("userevent", "Parse...")
            parsed_entry = self.prepParser(prep_entry)

            if parsed_entry:
                self.messages.message("userevent", "Fertig geparst")

                #Wir brauchen die ID und den Namen der neuen Prposition
                #ID als bergabeparamter fr die Berechnung,
                #Name zur Darstellung im Infofeld
                self.current_prep = parsed_entry[1]
                prep_name = self.prepcon.searchpreps(id = self.current_prep)
                prep_name = prep_name[0].name
                self.binfo_prep.configure(text = prep_name)

                #Fllung des bisherigen Objektes lschen
                self.canvas.interior().itemconfigure("ID"+self.karte.cobjects[self.current_object].id,
                                                     fill = "")
                #Mittels Id das richtige Objekt in der Liste der geladenen
                #Objekte finden und zum aktuellen Objekt machen
                self.current_object = 0
                max = len(self.karte.cobjects)
                while (self.karte.cobjects[self.current_object].id != parsed_entry[2][0])\
                      and (self.current_object < max):
                    self.current_object += 1

                #Objekt & Buffer zeichnen, Objektinfo anzeigen
                self.binfo_obj.configure(text = self.karte.cobjects[self.current_object].name)
                self.canvas.interior().itemconfigure("ID"+self.karte.cobjects[self.current_object].id,
                                                     fill = self.ofill)
                self.drawBuffer(self.karte.cobjects[self.current_object])
       

    #----------------------------------------------------------------Viewer
    #  smtliche Viewerbehandlung, inklusive Events, callback-handler etc.)

    #___________________________________________________Viewer-callback-handlers
    def setViewer(self):
        """Erlaubt dem User, die Viewer-Werte per Mausklick auf dem Canvas
           zu setzen."""

        #Alle alten Mausbindings merken, sollen wieder hergestellt werden,
        #nachdem der Viewer gesetzt oder rechts geklickt wurde.
        self.old_button1 = self.canvas.interior().bind("<Button-1>")
        self.old_button3 = self.canvas.interior().bind("<Button-3>")
        self.old_motion = self.canvas.interior().bind("<Motion>")

        #Die Maustasten werden neu belegt
        self.canvas.interior().bind('<Button-1>', self.onSetViewerClick)
        self.canvas.interior().bind('<Button-3>', self.onSetViewerRightClick)

        self.canvas.config(cursor="crosshair")


    #______________________________________________________________Viewer-Events
    def onSetViewerClick(self, event):
        """bernimmt die aktuellen Canvas-Koordinaten fr den Viewer
           und leitet das Setzen der Richtung ein."""

        #Wo steht die Maus? Zoom beachten!
        x = self.canvas.canvasx(event.x) / self.zoom
        y = self.canvas.canvasy(event.y) / self.zoom
        self.setViewerValues(coords = [x, y])

        if self.set_viewer_dir:
            #Es soll noch eine Viewer-Richtung gesetzt werden. Dazu ntig:
            #Neue Binds zur Einleitung des Richtungssetzens
            self.canvas.interior().bind('<Motion>', self.onSetViewerDirMove)
            self.canvas.interior().bind('<Button-1>', self.onSetViewerDirClick)
        else:
            #Keine Richtungseingabe erwnscht, also sind wir fertig.
            #Jetzt aufrumen (bindings, crosshair) und Werte anwenden
            self.canvas.interior().bind('<Button-1>', self.old_button1)
            self.canvas.interior().bind('<Button-3>', self.old_button3)
            self.canvas.config(cursor="")
            
            self.modifyViewer(state=ON)
            

    def onSetViewerDirClick(self, event):
        """Der Punkt, an dem links geklickt wurde, wird der neue Endpunkt
           der Viewer-Richtung."""

        #Die Richtungs-Setzen-Linie erstmal lschen
        self.canvas.delete("DirMoveLine")

        #Richtungsvektor = Differenz zwischen Klick und Viewer
        #(Zoom und Canvasumrechnung beachten)
        dirx = self.canvas.canvasx(event.x) / self.zoom
        diry = self.canvas.canvasy(event.y) / self.zoom
        dirx -= float(self.viewer_xcoord_entry.getvalue())
        diry -= float(self.viewer_ycoord_entry.getvalue())
        
        #Richtungs-Entry-Felder aktualisieren
        self.setViewerValues(dir = [dirx, diry])

        #Ursprngliche Mausbindings wieder herstellen
        self.canvas.interior().bind('<Button-1>', self.old_button1)
        self.canvas.interior().bind('<Motion>', self.old_motion)
        self.canvas.config(cursor="")

        #Neue Viewerwerte anwenden
        self.modifyViewer(state=ON)


    def onSetViewerDirMove(self, event):
        """berprft whrend des Viewer-Direction-Setzens fortwhrend, wo
           der Mauszeiger gerade ist und malt eine Linie dorthin."""
        
        #Damit der Benutzer auch gleich beim Setzen die Werte sieht:
        self.showCoordinates(event)
        
        #Viewerwerte und Mauskoordinaten rausfinden
        viewerx = float(self.viewer_xcoord_entry.getvalue())
        viewery = float(self.viewer_ycoord_entry.getvalue())
        dirx = self.canvas.canvasx(event.x) / self.zoom
        diry = self.canvas.canvasy(event.y) / self.zoom

        #Falls schon eine Linie gemacht wurde, bewege sie.
        #Ansonsten mach erstmal eine
        if self.canvas.coords("DirMoveLine"):
            self.canvas.coords("DirMoveLine",
                               viewerx*self.zoom, viewery*self.zoom,
                               dirx*self.zoom, diry*self.zoom
                               )
        else:
            self.canvas.create_line(viewerx*self.zoom, viewery*self.zoom,
                                    dirx*self.zoom, diry*self.zoom,
                                    fill="black",
                                    tag="DirMoveLine"
                                    )


    def onSetViewerRightClick(self, event):
        """Bricht das Setzen des Betrachters ab und stellt alle vorher
           gltigen Werte (Viewer und Maus-Bindings) wieder her. """

        self.canvas.delete("DirMoveLine")
        
        self.canvas.interior().bind('<Button-1>', self.old_button1)
        self.canvas.interior().bind('<Button-3>', self.old_button3)
        self.canvas.interior().bind('<Motion>', self.old_motion)
        self.canvas.config(cursor="")
        if self.viewer:
            self.viewer_xcoord_entry.setvalue(self.viewer.x)
            self.viewer_ycoord_entry.setvalue(self.viewer.y)
            if self.set_viewer_dir:
                self.viewer_xdir_entry.setvalue(self.viewer.direction[0])
                self.viewer_ydir_entry.setvalue(self.viewer.direction[1])
        else:
            self.viewer_xcoord_entry.setvalue(self.inactive_viewer.x)
            self.viewer_ycoord_entry.setvalue(self.inactive_viewer.y)
            if self.set_viewer_dir:
                self.viewer_xdir_entry.setvalue(self.inactive_viewer.direction[0])
                self.viewer_ydir_entry.setvalue(self.inactive_viewer.direction[1])
            
        
        
    #___________________________________________________weitere Viewerfunktionen
    def getViewerValues(self, name, id):
        """Holt die aktuellen Benutzerwerte aus einer der folgenden
           Quellen (nach Prioritt):
           1. Die Entry-Felder
           2. Die Koordinaten im aktiven Viewer
           3. Die Koordinaten aus dem inaktiven Viewer

           Bentigt <name> und <id> des Viewers als bergabewerte."""
           #Dass <name> und <id> bentigt werden ist eigentlich unschn und
           #resultiert alleine auf der ursprnglichen Idee, mehrere Viewer
           #zuzulassen...
        
        nviewer, dir, coords = None, None, []

        #**************************************************** x-Koordinaten
        if self.viewer_xcoord_entry.checkentry() == Pmw.OK:
            coords.append(float(self.viewer_xcoord_entry.getvalue()))
        elif self.viewer:
            coords.append(self.viewer.x)
            self.setViewerValues(coords = [self.viewer.x, None])
        else:
            coords.append(self.inactive_viewer.x)
            self.setViewerValues(coords = [self.inactive_viewer.x, None])
            
        #**************************************************** y-Koordinaten
        if self.viewer_ycoord_entry.checkentry() == Pmw.OK:
            coords.append(float(self.viewer_ycoord_entry.getvalue()))
        elif self.viewer:
            coords.append(self.viewer.y)
            self.setViewerValues(coords = [None, self.viewer.y])
        else:
            coords.append(self.inactive_viewer.y)
            self.setViewerValues(coords = [None, self.inactive_viewer.y])

        #********************************************************* Richtung
        if self.set_viewer_dir:
            dir = []
            dir.append(float(self.viewer_xdir_entry.getvalue()))
            dir.append(float(self.viewer_ydir_entry.getvalue()))
        
        #********************************************************* Rckgabe
        nviewer = ViewerObject(name, id, coords, dir)

        return nviewer        


    def modifyViewer(self, state = None, coords = None, dir = None):
        """ndert Werte oder Status des Viewers.
           Falls <state>, <coords> und/oder <dir> bei der bergabe
           Werte erhalten, besitzt der Viewer danach diese Werte, fehlende
           Werte werden aus den Entry-Feldern, Viewer-Objekt-Daten und dem
           Status-Switch (<self.viewer_state>) gewonnen.
        """

        if self.viewer:
            if self.viewer_state.get():
                if coords and (not state or state == ON):
                    #Viewer aktiv, programminterne Wertnderung 
                    self.setViewerValues(coords, dir)
                elif (not coords) and (not state or state == ON):
                #Viewer aktiv, Usereingabe in Entryfeldern oder
                #Viewer aktiv, Usereingabe durch Mausklick
                    pass
                #Jetzt stehen in allen bisher genannten Fllen die Werte im Entry
                self.viewer = self.getViewerValues(self.viewer.name, self.viewer.id)
            else:
                if not coords and state == ON:
                    #Nach Initialisierung:
                    #  viewer hat Werte,
                    #  keine Werte im Entry
                    #  viewer_state = 0 (also kein Hkchen gesetzt) 
                    self.setViewerValues(self.viewer.getXY(), self.viewer.direction)
                    self.viewer_state.set(ON)
                elif not coords and not state:
                    #Viewer aktiv, User hat deaktiviert (Checkbox)
                    self.canvas.delete(self.viewer.name+self.viewer.id)
                    self.inactive_viewer = copy.deepcopy(self.viewer)
                    self.viewer = None
        else:
            if self.viewer_state.get():
                #Viewer inaktiv, User hat aktiviert (Checkbox)
                self.viewer = copy.deepcopy(self.inactive_viewer)
                self.inactive_viewer = None
            else:
                if coords and state == ON:
                    #Viewer inaktiv, neuer Viewer durch Mausklick positioniert
                    self.inactive_viewer = None
                    self.setViewerValues(coords, dir)
                    self.viewer = self.getViewerValues(self.viewer.name, self.viewer.id)
                elif coords and not state:
                    #Viewer inaktiv, programminterne Wertnderung
                    self.setViewerValues(coords, dir)
                    self.inactive_viewer = self.getViewerValues(self.inactive_viewer.name, self.inactive_viewer.id)
                elif not coords and not state:
                    #Viewer inaktiv, keine Wertnderung, bleibt inaktiv
                    pass
                elif not coords and state == ON:
                    #Viewer inaktiv, Usereingabe in Entryfeldern
                    self.viewer = self.getViewerValues(self.inactive_viewer.name, self.inactive_viewer.id)
                    self.inactive_viewer = None
                else:
                    self.showError(InternalError("bgui.GUI.modifyViewer", 1))

        if self.viewer:
            self.viewer_state.set(ON)
            self.drawViewer()

        #Egal, wie der Viewer verndert wurde, falls ein Buffer berechnet wurde,
        #muss er jetzt neu berechnet werden
        if self.showing_buffer:
            self.drawBuffer(self.karte.cobjects[self.current_object])


    def setViewerValues(self, coords = [None, None], dir = [None, None]):
        """Schreibt neue Werte in Viewer-Entries.
           Koordinaten <coords> und Richtung <dir> des Viewers werden in
           2er-Listen bergeben. Alle nicht besetzten Werte werden auch
           nicht gendert."""

        if coords[0]: self.viewer_xcoord_entry.setvalue(round(coords[0],1))
        if coords[1]: self.viewer_ycoord_entry.setvalue(round(coords[1],1))
        if self.set_viewer_dir and dir:
            if dir[0]: self.viewer_xdir_entry.setvalue(round(dir[0],1))
            if dir[1]: self.viewer_ydir_entry.setvalue(round(dir[1],1))


    #--------------------------------------------------------sonstige-Funktionen
    def centerOnObject(self, obj):
        """Scrollt den canvas so, dass das Objekt <obj> in der Mitte steht.
        """
        
        #Wo liegt (ca.) die Mitte des Objekts?
        objbb = BBgeo(obj)
        obj_center = objbb.getcenter()
        
        #Aufruf der eigentlich Scroll-Methode
        self.centerOnXY(obj_center.x, obj_center.y)


    def centerOnXY(self, x, y):
        """Scrollt den Canvas so, dass der Punkt x/y in der Mitte steht.
        """
        
        #X-Offset rausfinden
        xview = self.canvas.xview()
        xoff = x / self.karte.width()
        xoff = xoff - ((xview[1] - xview[0]) / 2)

        #Y-Offset rausfinden
        yview = self.canvas.yview()
        yoff = y / self.karte.height()
        yoff = yoff - ((yview[1] - yview[0]) / 2)

        #Canvas verschieben
        self.canvas.xview(MOVETO,xoff)
        self.canvas.yview(MOVETO,yoff)


    def cycleObjects(self, step):
        """Whlt ein neues Kartenobject, das in der Liste <step> Eintrge
           entfernt vom aktuellen Eintrag ist, und setzt die Objektfllung
           entsprechend um.
        """

        #Objektfllung wird hier behandelt, da nach der Neuwahl das alte
        #Objekt nicht mehr bekannt ist
        self.canvas.interior().itemconfigure("ID"+self.karte.cobjects[self.current_object].id,
                                             fill = "")
        self.current_object = btools.indexcycle(self.current_object + step, len(self.karte.cobjects))
        self.binfo_obj.configure(text = self.karte.cobjects[self.current_object].name)
        self.canvas.interior().itemconfigure("ID"+self.karte.cobjects[self.current_object].id,
                                             fill = self.ofill)

        self.drawBuffer(self.karte.cobjects[self.current_object])


    def drawBuffer(self, object):     
        """Berechnet und zeichnet einen Buffer.
           Der Buffer zu <object> wird unter Bercksichtigung der gerade
           gltigen Prposition und des gerade gltigen Viewers berechnet.

           <object> muss vom Typ GeoObject (aus geoobject.py) sein.

           Es wird mindestens der Buffer gezeichnet, eventuell auch die
           BBs.

           Es gibt keinen Rckgabewert.
        """

        #self.showing_buffer ist ein Flag dafr, dass _versucht_ wurde,
        #einen Buffer anzuzeigen.
        self.showing_buffer = 1

        #............................................alte Buffer und BBs lschen
        self.canvas.delete("bbox_all")
        self.canvas.delete("buffer")

        #.......................................................Buffer berechnen
        try:
            buffer=Bufferlo([object.id], self.objdb,
                                 self.current_prep, self.prepdb,
                                 self.viewer, None)
        except (GeometryError, ParameterError, SemanticError, DataError), fehler:
            self.showError(fehler)            
        else:
            #................................................Default-BB zeichnen
            punkte, zaehler = [], 0
            for point in buffer.buffer[2].points:
                point = point.mult(self.zoom)
                punkte.append(point.getXY())

            #.................................BB-Polygon + Beschriftung zeichnen
            #Falls die BB gezeichnet werden soll, sollen auch die Linien
            #nummeriert werden
            if self.dbb_line:
                textcolor = "black"
            else:
                textcolor = ""
            self.canvas.create_polygon(punkte,
                                       fill="", outline=self.dbb_line,
                                       tag=("dbbox", "bbox_all")
                                       )
            #An jeden einzelnen Eckpunkt der BB die Linienummer der dort
            #startenden BB-Linie
            for zaehler in range(len(buffer.buffer[2].lines)):
                point = copy.deepcopy(buffer.buffer[2].lines[zaehler].P1)
                point = point.mult(self.zoom)
                self.canvas.create_text(point.getXY(), fill = textcolor,
                                        text=zaehler+1,
                                        tag=("dbbox-text", "bbox_all")
                                        )
            
            #.....................................richtungsabhngige BB zeichnen
            punkte, zaehler = [], 0
            for point in buffer.buffer[1].points:
                point = point.mult(self.zoom)
                punkte.append(point.getXY())
            #BB-Polygon + Beschriftung zeichnen        
            if self.bb_line:
                textcolor = "black"
            else:
                textcolor = ""
            self.canvas.create_polygon(punkte,
                                       fill="", outline=self.bb_line,
                                       tag=("bbox", "bbox_all")
                                       )
            for zaehler in range(len(buffer.buffer[1].lines)):
                point = copy.deepcopy(buffer.buffer[1].lines[zaehler].P1)
                point = point.mult(self.zoom)
                self.canvas.create_text(point.getXY(), text=zaehler+1,
                                        fill = textcolor,
                                        tag=("bbox-text", "bbox_all")
                                        )
            self.canvas.interior().lower("bbox")
            self.canvas.interior().lower("bbox-text")
            self.canvas.interior().lower("dbbox")
            self.canvas.interior().lower("dbbox-text")
                
            #..............................................neuen Buffer zeichnen
            for x in buffer.buffer[0]:
                punkte = []
                for point in x.points:
                    point = point.mult(self.zoom)
                    punkte += point.getXY()
                self.canvas.create_polygon(punkte,
                                           fill = self.bfill, outline = "blue",
                                           tag="buffer"
                                           )
                self.canvas.interior().lower("buffer")
            self.centerOnObject(object)


    def drawViewer(self, size=20):
        """Zeichnet Viewer [+ Direction].
        
           Der Viewer wird als Kreuz gezeichnet, dessen Striche die Lnge <size>
           haben. Die Direction wird als Pfeil gezeichnet, der mindestens die
           Lnge <size>*2 hat.
           
           Es gibt keinen Rckgabewert.
        """

        #Alten Viewer von der Karte loeschen
        if self.canvas.find_withtag(self.viewer.name+self.viewer.id):
            self.canvas.delete(self.viewer.name+self.viewer.id)
        
        #Betrachter-Kreuz zeichnen
        x = self.viewer.x * self.zoom
        y = self.viewer.y * self.zoom
        x1 = x - (size/2)
        x2 = x + (size/2)
        y1 = y + (size/2)
        y2 = y - (size/2)
        self.canvas.create_line(x1,y,x2,y,
                                fill="green",
                                width = 2,
                                tag=self.viewer.name+self.viewer.id
                                )
        self.canvas.create_line(x,y1,x,y2,
                                fill="darkgreen",
                                width = 2,
                                tag=self.viewer.name+self.viewer.id
                                )

        #Richtungspfeil zeichnen
        if self.set_viewer_dir:
            x1 = x + (self.viewer.direction[0] * self.zoom)
            y1 = y + (self.viewer.direction[1] * self.zoom)
            dir_line=Line(None,Point(None,x,y),Point(None,x1,y1))
            if dir_line.length() < size * 2: dir_line.scaleToLength(size*2)
        
            self.canvas.create_line(dir_line.getXY(),
                                    fill="blue",
                                    tag=self.viewer.name+self.viewer.id,
                                    arrow=LAST)   

        
    def loadMap(self, filename, filetype):
        """Ldt ein neues Kartenobjekt.
        
           <filename>: Dateiname der zu ladenden Karte
           <filetype>: Datentyp der Datei, momentam wird nur XML untersttzt.

           Erfolg: <self.karte> (Instanz von GeoMap) enthlt neue Karte
                   <self.objdb> enthlt den neuen Dateinamen
                   
           Mierfolg: Keine nderungen in den Werten, entsprechende
           Fehlermeldungen.
           
           Es gibt keinen Rckgabewert.
        """


        if os.access(filename, os.F_OK):
            self.messages.message("systemevent", "Lese neue Karte...")
            neue_karte = GeoMap(filename, XML)
            try:
                neue_karte.getparams()
            except (ParameterError, SemanticError), fehler:
                self.showError(fehler)
            else:
                try:
                    neue_karte.getobjects()
                except ParameterError, fehler:
                    self.showError(fehler)
                else:
                    self.objdb = filename
                    self.karte = copy.deepcopy(neue_karte)
                    self.updateMap(mode = NEW)
        else:
            self.showError(GuiError("bgui.GUI.loadMap", 1))
            

    def loadPreps(self, filename, filetype):
        """Ldt eine neue Prpositionsdatenbank.

           <filename>: Dateiname der zu ladenden Datenbank
           <filetype>: Datentyp der Datei, momentan nur XML

           Erfolg: <self.prepdb> erhlt <filename>
                   <self.preptyp> erhlt <filetyp>
                   <self.current_prep> zeigt auf 1. Prp aus der Datenbank

           Mierfolg: Keine nderungen in den Werten, entsprechende
           Fehlermeldungen.
           
           Es gibt keinen Rckgabewert.          
        """


        self.prepcon = prepcontainer.PrepContainer(filename, filetype)

        try:
            self.prepcon.getpreps(projection = self.karte.pov)
        except ParameterError, fehler:
            self.showError(fehler)
        else:
            self.current_prep = self.prepcon.cpreps[0].id            
            self.binfo_prep.configure(text = self.prepcon.cpreps[0].name)
            self.prepdb = filename
            self.preptyp = filetype


    def loadGram(self, filename, filetype):
        """Ldt eine neue Grammatikdatei.

           <filename>: Dateiname der zu ladenden Datenbank
           <filetype>: Datentyp der Datei, momentan nur XML

           Erfolg: <self.gramdb> erhlt <filename>
                   <self.gramtyp> erhlt <filetyp>
                   <self.current_prep> zeigt auf 1. Prp aus der Datenbank

           Mierfolg: Keine nderungen in den Werten, entsprechende
           Fehlermeldungen.
           
           Es gibt keinen Rckgabewert."""

        if os.access(filename, os.F_OK):
            self.gramdb = filename
            self.gramtyp = filetype
        else:
            self.showError(GuiError("bgui.GUI.loadGram", 1))


    def prepParser(self, entry):
        """Parst einen String mit dem PTag-Parser.

           <entry> ist der zu parsende String,
           weitere bergabewerte fr die PTag-Instanz werden aus den
           entsprechenden Gui-Objekt-Daten geholt.

           Erfolg: Es wird eine Liste mit
                   [[mod1, mod2, ...], Prep, [obj, obj2, ...]]
                   zurckgegeben.
           Mierfolg: Es wird None zurckgegeben
        """

        ergebnis = None
        try:
            parsed = ptag.PTag(entry,
                          self.prepdb,
                          self.preptyp,
                          self.objdb,
                          self.objtyp,
                          self.gramdb,
                          self.gramtyp)
        except (ParserError, DataError), fehler:
            self.showError(fehler)
        else:
            #Falls Parserergebnis nicht eindeutig:
            #  gib Warnung aus, mach mit 1. Ergebnis weiter
            if len(parsed.results) > 1:
                self.showError(ParserError("bgui.prepParser",4), type = WARNING)
            ergebnis = parsed.results[0]            
        
        return ergebnis


    def showCopyRight(self):
        """Ruft die CopyRight-Klasse auf."""
        cr = CopyRight()


    def showError(self, fehler):
        """Schreibt eine Fehlermeldung in ein (neues) Fenster."""
        if not self.error_window:
            self.error_window = BGuiError()
        self.error_window.showError(fehler)


    def showHelp(Self):
        """Ruft die Help-Klasse auf."""
        help = Help()


    def updateMap(self, mode = NEW):
        """Zeichnet die komplette Karte im Canvas, aktualisiert Karteninfo.
        
           <mode> bestimmt, ob das aktuelle Objekt, eventuell gezeichnete Buffer
                  und die Karteninfos stehen bleiben oder resettet werden.
        """

        #....................................................Alte Karte lschen
        self.canvas.delete("all")

        #.....................................................Objekte darstellen
        punkte = []

        #Erstmal die jeweiligen Objektpunkte sammeln
        for i in self.karte.cobjects:
            for point in i.points:
                point = point.mult(self.zoom)
                punkte.append(point.getXY())

            #Jetzt zeichnen und beschriften
            self.canvas.create_polygon(punkte,
                                       fill="", outline="red",
                                       tag="ID"+i.id)
            self.canvas.create_text(punkte[0], text=i.name)
            self.balloon.tagbind(self.canvas.interior(), "ID"+i.id,
                                 "Name: " + i.name+"\nID: " + i.id
                                 )
                                 
            punkte = []

            #Richtungslinien zum aktuellen Objekt zeichnen
            #kein Polygon, sondern mehrere Linien
            if i.direction:
                punkte = i.direction.getPoints()
                punkte_copy = copy.deepcopy(punkte)
                for point in range(len(punkte)):
                    punkte_copy[point] = punkte_copy[point].mult(self.zoom)

            linie = 0
            while linie < len(punkte_copy)-1:
                self.canvas.create_line(punkte_copy[linie].getXY(),
                                        punkte_copy[linie+1].getXY(),
                                        arrow = LAST,
                                        fill = "darkblue",
                                        tag = "dirarrow")
                linie += 1

            punkte=[]
        self.canvas.resizescrollregion()

        #.............................................Reset der Anzeigeparameter
        if mode == NEW:
            self.modifyViewer()

            self.current_object = 0
            self.binfo_obj.configure(text = self.karte.cobjects[self.current_object].name)
            self.canvas.interior().itemconfigure("ID"+self.karte.cobjects[self.current_object].id,
                                                 fill = self.ofill)
            if not self.current_prep:
                self.loadPreps(self.prepdb, self.preptyp)
            # Die neuen Kartenparameter ins Kartenfeld schreiben
            karteninfo = "KARTE: "+ str(self.karte.name)\
                         + "\nAnsicht: " + str(self.karte.pov)\
                         + "\nNord-Sd: " + str(self.karte.maxnorth) + " - "\
                         + str(self.karte.maxsouth) + "\nWest-Ost: " \
                         + str(self.karte.maxwest) + " - "\
                         + str(self.karte.maxeast)
            self.mapinfo.config(text=karteninfo)
            self.messages.message("systemevent", "Neue Karte gelesen")
        elif mode == REDRAW:
            self.messages.message("systemevent", "Karte neu gezeichnet")
            self.canvas.interior().itemconfigure("ID"+self.karte.cobjects[self.current_object].id,
                                                 fill = self.ofill)
            self.drawBuffer(self.karte.cobjects[self.current_object])
            self.modifyViewer()

        
#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
class BGuiError:
    """Gibt in einem seperaten Fenster eine Fehlermeldung aus.

       Verwendung:
       Vor der ersten Verwendung ohne Parameter instanziieren. Zur tatschlichen
       Fehlerdarstellung mu dann <showError>, nheres siehe dort.
       
       Alle anderen Funktionen sind intern.       
    """

    def __init__(self):
        """Hier wird das Fenster initialisiert, labels gesetzt usw.
           Es werden keinerlei zustzliche Parameter bentigt.
        """
        
        #Erstmal ein paar Counter setzen
        self.error_list = [] #Liste mit noch nicht abgesegneten Fehlern
        self.error_count = 0 #Gesamtfehleranzahl whrend Programmlaufzeit
        self.open = 0        #Fehlerfenster gerade offen?

        #Allgemeine Fenstereigenschaften setzen
        self.error_win = Toplevel()
        self.error_win.title("Bufferlo Fehler")
        self.error_win.protocol("WM_DELETE_WINDOW", self._hideError)

        #Widgets setzen
        #Fehlername (mit Fehlernummer)
        self.error_name = Label(self.error_win)
        self.error_name.pack(anchor = W, expand = YES)
        #Die eigentliche Fehlermeldung
        self.error_message = Pmw.ScrolledField(self.error_win, entry_bd = 1)
        self.error_message.pack(expand = YES, fill = X)
        #Erklrung mglicher Fehlerursachen
        self.error_cause = Pmw.ScrolledText(self.error_win,
                                       labelpos = NW,
                                       label_text = "Ursache",
                                       text_width = 60,
                                       text_height = 7,
                                       text_bd = 1,
                                       text_bg = None,
                                       text_wrap = "word",
                                       text_state = "disabled",
                                       )
        self.error_cause.pack(expand = YES, fill = X, pady = 2)
        #Auf welche Datenquelle wurde zugegriffen (welches file)
        self.error_data_source = Pmw.ScrolledField(self.error_win,
                                                   labelpos = NW,
                                                   label_text = "Datenquelle",
                                                   entry_bd = 1)
        self.error_data_source.pack(expand = YES, fill = X, pady = 2)
        #Nach welchen Daten wurde gesucht
        self.error_requested_data = Pmw.ScrolledField(self.error_win,
                                                      labelpos = NW,
                                                      label_text = "Angeforderte Daten",
                                                      entry_bd = 1)
        self.error_requested_data.pack(expand=YES, fill=X, pady=2)

        #Und noch Knpfe
        self.error_more_errors = Button(self.error_win,
                                        text = "Keine weiteren Fehler",
                                        state = "disabled",
                                        command = (lambda: self._nextError()))
        self.error_more_errors.pack(side = LEFT)
        self.error_ignore = Button(self.error_win,
                                   text = "Ignoriere (weitere) Fehler",
                                   command = (lambda: self._hideError()))
        self.error_ignore.pack(side = LEFT)


    def showError(self, error_object):
        """Entscheidet, ob ein neuer Fehler gezeigt wird, oder in die
           Fehlerliste kommt.
           Falls der Fehler angezeigt wird, wird <_displayError> mit
           <error_object> als Parameter aufgerufen, ansonsten kommt der Fehler
           in die Warteliste <error_list>.
           
           <error_object>: Muss vom Typ BError sein.
        """
        
        #Falls ein Fenster offen ist, kommt der neue Fehler ans Ende der Reihe
        if self.open:
            self.error_list.append(error_object)
        else:
            self.open = 1
            self._displayError(error_object)

        #Fenster evt. wiederherstellen
        if self.error_win.state() == "withdrawn":
            self.error_win.deiconify()

        self._updateButtons()
        
        #Und jetzt in den Vordergrund
        self.error_win.tkraise()        


    def _displayError(self, fehler):
        """Interne Funktion.
           Zeigt einen Fehler im separaten Fehlerfenster an.
           Auerdem wird die Gesamtfehleranzahl <error_count> mitgezhlt.

           <fehler> muss vom Typ BError sein.            
        """
        
        self.error_count = self.error_count + 1
        #Fehlernummer + Name sollen zusammen angezeigt werden
        name_text = "#" + str(self.error_count) + ": " + fehler.name
        if DEBUG:
            name_text += " (aus " + fehler.ort + ")"

        #Die einzelnen Ausgaben sind in __init__ erklrt
        self.error_name.configure(text = name_text)
        self.error_message.configure(text = fehler.meldung)
        self.error_cause.setvalue(fehler.ursache)
        self.error_data_source.configure(text = "")
        self.error_requested_data.configure(text = "")
        
        #Nur bei Fehlern der Art DataError mssen die beiden letzten Ausgaben
        #gefllt werden
        if isinstance(fehler, DataError):
            if fehler.data_source:
                self.error_data_source.configure(text = fehler.data_source)
            if fehler.requested_data:
                self.error_requested_data.configure(text = fehler.data_source)


    def _updateButtons(self):
        """Passt den Meldungstext im Ignore-Button den noch vorhandenen
           Fehlern an.
        """
        
        error_count = len(self.error_list)
        if error_count == 0:
            self.error_more_errors.config(text = "Keine weiteren Fehler",
                                          state = "disabled")
            self.error_ignore.config(text = "Ignoriere Fehler")
        elif error_count == 1:
            self.error_more_errors.config(text = "1 weiterer Fehler",
                                          state = "active")
            self.error_ignore.config(text = "Ignoriere weitere Fehler")
        else:
            self.error_more_errors.config(text = str(error_count) + " weitere Fehler")


    def _nextError(self):
        """Lscht den aktuell dargestellten Fehler aus der Liste und zeigt den
           nchsten an.
        """
        
        self._displayError(self.error_list[0])
        del self.error_list[0]
        self._updateButtons()
        

    def _hideError(self):
        """Das Fehlerfenster soll verschwinden.
           Vorher werden alle noch ausstehenden Fehler zur Fehleranzahl gezhlt.
        """
        
        self.error_count = self.error_count + len(self.error_list)
        self.error_list = []
        self.error_win.withdraw()
        self.open = 0


#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
class Help:
    """Die Klasse Help zeigt Tastenbelegung und Farblegende der Klasse Gui
       in einem separaten Fenster. Es ist eine sehr rudimentre Klasse, da
       berhaupt nicht variabel o...

       Der Aufruf erfolgt einfach ber eine Objektinstanziierung ohne weitere
       Parameter.               
    """
    
    def __init__(self):
        """Hier geschieht die komplette Hilfeanzeige, es werden keine Parameter
           bentigt.
        """
        
        win = Toplevel()
        win.title("Tastaturbelegung und Farblegende")
        help_text = Text(win,
                         height = 20,
                         width = 80,
                         bg = "#FFFFF8")
        help_text.insert(END, "TASTENBELEGUNG\n\nF1: Diese Hilfe\n"\
                         + "F2: Fllen des Buffers an/aus\n"\
                         + "F3: Vorheriges Kartenobjekt\n"\
                         + "F4: Nchstes Kartenobjekt\n"\
                         + "F5: Karte aktualisieren\n"\
                         + "F6: sichtweisenabhngige Boundingbox an/aus\n"\
                         + "F7: sichtweisenunabhngige Boundingbox an/aus\n"\
                         + "F8: Aktuelles Objekt fllen an/aus\n\n"\
                         + "FARB-LEGENDE\n\n"
                         + "rot: Kartenobjekte\n"
                         + "grn: sichtweisenabhngige Boundingbox\n"
                         + "pink: sichtweisenunabhngige Boundingbox\n"\
                         + "blau: Buffer\n"
                         + "schwarzer Text: Beschriftung\n"\
                         + "schwarze Zahlen: Nummerierung der BB-Seiten\n"\
                         + "blaue Pfeile: Richtungspfeile der Kartenobjekte\n"\
                         + "grnes Kreuz: Betrachter, falls aktiv"
                         )
        help_text.tag_add("rot", "14.0", "14.3")
        help_text.tag_add("gruen", "15.0", "15.4")
        help_text.tag_add("pink", "16.0", "16.4")
        help_text.tag_add("blau", "17.0", "17.4")
        help_text.tag_add("blau", "20.0", "20.12")
        help_text.tag_add("gruen", "21.0", "21.13")
        help_text.tag_config("rot", foreground="red")
        help_text.tag_config("gruen", foreground="green")
        help_text.tag_config("pink", foreground="pink")
        help_text.tag_config("blau", foreground="blue")
        help_text.pack()
        win.mainloop()


#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
class CopyRight:
    """Diese Klasse zeigt Copyright-Informationen in einem separaten Fenster.
       Dazu muss sie lediglich ohne weitere Parameter instanziiert werden.
    """

    def __init__(self):
        """Erstellt das Fenster und ruft die kurze Copyright-Notiz auf."""
        
        win = Toplevel()
        win.title("Copyright")
        self.cr_text = Pmw.ScrolledText(win,
                                        text_bg = "#FFFFF8",
                                        text_width = 80,
                                        text_wrap = "word",
                                        text_state = DISABLED)
        self.cr_text.pack()
        self.cr_button = Button(win)
        self.cr_button.pack()
        self.shortVersion()     
        
        win.mainloop()
        
    
    def shortVersion(self):
        """Zeigt die kurze Copyright-Notiz im Copyrightfenster."""
        
        copyright = "BUFFERLO\nCopyright (C) 2003 Verleih, Reichelt, "\
                    + "Heiob, Feiereisen\n\nContact: "\
                    + "Etienne Verleih: verleih@cl.uni-heidelberg.de\n"\
                    + "Thorsten Reichelt: Thorsten.Reichelt@web.de\n\n"\
                    + "Bufferlo comes with ABSOLUTELY NO WARRANTY. This "\
                    + "is free software, and you are welcome to "\
                    + "redistribute it under certain conditions; click "\
                    + "'TERMS' for further details.\n\n\n"\
                    + "geometry.py\nCopyright (C) 2001 von der Malsburg, "\
                    + "Copyright (C) 2002 Verleih, Reichelt\n\n"\
                    + "geobuffer.py\nCopyright (C) 2001 von der Malsburg\n\n\n"\
                    + "Thanks to Jan-Ole Esleben (Parser Prototype)"
        
        self.cr_text.setvalue(copyright)
        self.cr_button.configure(text = "TERMS")
        self.cr_button.configure(command = lambda: self.longVersion())
        

    def longVersion(self):
        """Zeigt die lange Copyright-Notiz im Copyrightfenster."""
        
        if os.access("COPYRIGHT", os.F_OK):
            datei = open("COPYRIGHT")
            copyright = datei.read()
            datei.close()
        else:
            copyright = "WARNING: File 'COPYRIGHT' not found. This file "\
                        + "should exist and contain the GNU General "\
                        + "Public License terms."
        self.cr_text.setvalue(copyright)
        self.cr_button.configure(text = "BACK")
        self.cr_button.configure(command = lambda: self.shortVersion())
        

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
        
#MAIN
userini = "user.ini"
if len(sys.argv) > 1:
    for argument in sys.argv[1:]:
        if argument[:2] == "-i":
            #alternative Ini-Datei
            userini = argument[2:]
        elif argument[:2] == "-d":
            #nur default-Ini
            userini = "default.ini"
        else:
            userini = None
            print """BGui, Copyright (C) 2003 Thorsten Reichelt
Aufruf: bgui.py [-i<inifile> | -d]

-i        ldt eine alternative Einstellungsdatei
inifile   Datei mit BGui-Einstellungen
-d        ldt BGui ohne Benutzereinstellungen"""


if userini:
    gui = Gui(userini)