# ****************************************************
# Name des Moduls: Btools
# Name des Projekts: Bufferlo
#
# Autor(en) des Moduls:
#       Thorsten Reichelt, Thorsten.Reichelt@web.de
#       Etienne Verleih, verleih@cl.uni-heidelberg.de
#
# Aufgaben des Moduls: siehe DocString
# Compiler fr eingebettete Dokumentation: PyDoc
#
# Datum der letzten nderung: 09.01.2003
# ****************************************************

""" Dieses Modul enthlt eine Sammlung von Funktionen zur String- und Listenverarbeitung.
    Diese sind im Einzelnen:

    - str_equals(): Vergleicht zwei Strings auf Gleichheit. Dafr gibt es eine eigene
    Funktion, um projektweit festlegen zu knnen, ob Stringvergeliche casesensitiv
    sein sollen oder nicht.

    - strlist_equals(): Vergleicht zwei Listen von Strings auf Gleichheit, benutzt dafr
    str_equals()

    - listcontains(): Gibt wahr zurck, wenn eine Liste einen String enthlt. Benutzt
    ebenfalls str_equals und soll das (x in y) von Python ersetzten.

    - indexcycle(): Errechnet einen neuen Index fr eine als rundlaufend angenommene
    Liste.

    - listrotate(): Macht das mit einem Index bezeichnete Element zum ersten Element der
    Liste, ohne die relative Reihenfolge der Elemente zu verndern.
"""

def str_equals(str1, str2):
    """ Diese Funktion gibt 1 zurck, wenn die beiden Strings gleich sind,
        0, wenn nicht. Die Funktion wird verwendet, um einheitlich zu regeln,
        ob Gross- Kleinschreibung beim Stringvergleich eine Rolle spielt
    """

    cmp = 0
    if str1.upper() == str2.upper():
        cmp = 1

    return cmp
       

def strlist_equals(l1, l2):
    """ berprft, ob zwei Listen mit Strings gleich sind (in gleicher Reihenfolge die
        gleichen Elemente enthalten). Benutzt dazu str_equals
    """
    
    cmp = 0
    len1 = len(l1)

    if len1 == len(l2):
        cmp = 1
        for i in range(len1):
            if not str_equals(l1[i], l2[i]):
                cmp = 0
                break;

    return cmp    


def listcontains(list, str):
    """ berprft, ob str gleich einem der Elemente aus list ist. Die Funktion
        soll die (x in y)-Funktion von Python ersetzen und zum Vergleich str_equals()
        verwenden. Gibt 1 zurck, wenn das Element enthalten ist, 0 sonst.
   """

    cmp= 0
    for x in list:
        if str_equals(x, str):
            cmp= 1

    return cmp    


def indexcycle(ziel, laenge):
    """ Berechnet eine neue Indexnummer in einer als rundlaufend angenommenen Liste.
        In ziel bergibt man die Position der Liste, die man erreichen mchte - diese
        Zahl kann beliebig weit ber der eigentlichen Lnge der Liste liegen. In laenge
        wird die tatschliche Lnge der Liste bergeben. Zurckgeliefert wird der Index
        des Elements, das erreicht wird, wenn die Liste als rundlaufend gedacht wird.
    """

    return ziel % laenge

        
def listrotate(list, index):
    """ Macht das Element an Position index der Liste list zum ersten Element der
        Liste, ohne die relative Reihenfolge der Listenelemente zu verndern
    """
    
    nlist = []
    l = len(list)
    for x in range(l):
        pos = indexcycle(index+x, l)
        nlist.append(list[pos])

    return nlist
