# ****************************************************
# Name des Moduls: Geobuffer
# Name des Projekts: Bufferlo
#
# Autor(en) des Moduls:
#       Titus von der Malsburg
#
# Aufgaben des Moduls: siehe DocString
# Compiler fr eingebettete Dokumentation: PyDoc
#
# Datum der letzten nderung: 12.01.2003
# ****************************************************

""" Diese Bibliothek wurde uns (ein groes Dankeschn!!) von Titus
    von der Malsburg zur Verfgung gestellt, der sie fr ein Programm
    im Rahmen eines Seminars ber Sprachgenerierung entwickelt hat.
    Im Augenblick wird aus dem Modul nur eine einzige Funktion
    verwendet: getBoundingBox() (von BBox, siehe auch dort). Da
    jedoch die restliche Funktionalitt vielleicht fr Erweiterungen
    von Bufferlo noch wichtig werden knnte, wollten wir auf jeden
    Fall das ganze Modul mitgeben.
"""


#####
# Titus' kleine Bibliothek mit der man Buffer und Bereiche
# zu Objekten der Bibliothek geometry machen kann
#####

import math, copy, geometry

def buffer(obj, d):
    """ Ruft je nach dem was die Argumente sind die entspechende Methode zur Generierung eines Buffers auf."""
    if isinstance(obj, geometry.Point):
        return PointBuffer(obj,d)
    elif isinstance(obj, geometry.Line):
        return LineBuffer(obj,d)
    elif isinstance(obj, geometry.Path):
        return PathBuffer(obj,d)
    elif isinstance(obj, geometry.Polygon):
        return PolygonBuffer(obj,d)
    elif isinstance(obj, geometry.Circle):
        return CircleBuffer(obj,d)

def PointBuffer(p,d):
    """ Macht einen Buffer zu einem Point (Circle) """
    if d < 0:
        raise "The area of the requested buffer would be less than 0!"
    return geometry.Circle(None,p,d)

def LineBuffer(ln,d):
    """ Liefert Buffer zu einer Linie (Polygon) """
    if d < 0:
        raise "The area of the requested buffer would be less than 0!"
    ln1 = copy.deepcopy(ln)
    ln1.scaleToLength(ln1.length()+2*d)
    
    lnt = copy.deepcopy(ln)
    lnt.scaleToLength(d)
    tp = geometry.Point(None,0,0)-lnt.P1
    lnt.translate(tp)
    lnt.rotate(90)
    tp = lnt.P2

    ln2 = copy.deepcopy(ln1)
    ln1.translate(tp)
    ln2.translate(geometry.Point(None,0,0)-tp)

    tp.rotate(90)
    ln1.translate(tp)
    ln2.translate(tp)

    return geometry.Polygon(None, [ln1.P1,ln1.P2,ln2.P2,ln2.P1])

def PathBuffer(p,d):
    """ Liefert den Buffer zu einem Path (Polygon) """
    if d < 0:
        raise "The area of the requested buffer would be less than 0!"
    lines1 = []
    lines2 = []
    for ln in p.getLines():
        ln1 = copy.deepcopy(ln)
        ln1.scaleToLength(ln1.length()+2*d)
        
        lnt = copy.deepcopy(ln)
        lnt.scaleToLength(d)
        tp = geometry.Point(None,0,0)-lnt.P1
        lnt.translate(tp)
        lnt.rotate(90)
        tp = lnt.P2
        
        ln2 = copy.deepcopy(ln1)
        ln1.translate(tp)
        ln2.translate(geometry.Point(None,0,0)-tp)

        tp.rotate(90)
        ln1.translate(tp)
        ln2.translate(tp)

        lines1.append(ln1)
        lines2.append(ln2)

    lines2.reverse()
    for ln in lines2:
        ln.reverse()
    lines = lines1 + lines2

    k = 0
    buf = []
    i = len(lines)/2
    buf.append(lines[0].P1)
    while k < len(lines)-1:
        if k == i-1:
            buf.append(lines[k].P2)
            buf.append(lines[k+1].P1)
        elif not geometry.collinear(lines[k].P1,lines[k].P2,lines[k+1].P2):
            buf.append(geometry.intPoint(lines[k],lines[k+1]))
        k = k+1
    buf.append(lines[-1].P2)
    return geometry.Polygon(None, buf)

def PolygonBuffer(p,d):
    """ Liefert den Buffer zu einem Polygon (Polygon) """
    lines1 = []
    lines2 = []
    for ln in p.getLines():
        ln1 = copy.deepcopy(ln)
        ln1.scaleToLength(ln1.length()+2*d)
        
        lnt = copy.deepcopy(ln)
        lnt.scaleToLength(d)
        tp = geometry.Point(None,0,0)-lnt.P1
        lnt.translate(tp)
        lnt.rotate(90)
        tp = lnt.P2
        
        ln2 = copy.deepcopy(ln1)
        ln1.translate(tp)
        ln2.translate(geometry.Point(None,0,0)-tp)

        tp.rotate(90)
        ln1.translate(tp)
        ln2.translate(tp)

        lines1.append(ln1)
        lines2.append(ln2)

    buffers = []
    for lines in [lines1,lines2]:
        k = 0
        buf = []
        while k < len(lines)-1:
            if not geometry.collinear(lines[k].P1,lines[k].P2,lines[k+1].P2):
                buf.append(geometry.intPoint(lines[k],lines[k+1]))
            k = k+1
        buf.append(geometry.intPoint(lines[0],lines[-1]))
        buffers.append(geometry.Polygon(None, buf))

    if geometry.contains(buffers[0],buffers[1]):
        return buffers[1]
    elif geometry.contains(buffers[1],buffers[0]):
        return buffers[0]
    else:
        raise """Etwas schreckliches ist passiert:\12Es konnte kein Buffer erzeugt werden, da das Ausgangspolygon sich selbst scheidet."""

def CircleBuffer(c,d):
    """ Liefert einen Buffer zu einem Circle (Circle)"""
    if c.getRadius()+d < 0:
        raise "The area of the requested buffer would be less than 0!"
    return geometry.Circle(None,c.getCenter(),c.getRadius()+d)

def fprint(objs):
    """ gibt eine liste von polygonen als gnuplotdata aus """
    fp = open('plot.dat','w')
    list = range(100)
    i=0
    while i<len(list):
        list[i]=range(len(objs))
        i=i+1
    k=0
    while k < len(objs):
        i=0
        while i < len(objs[k].getPoints()):
            list[i][k] = (objs[k].getPoints())[i]
            i=i+1
        k=k+1
    i=0
    for ln in list:
        for p in ln:
            if not isinstance(p, geometry.Point):
                continue
            fp.write("%s %s "%(p.x,p.y))
        fp.write("\n")
    fp.close()

def getBoundingBox(obj):
    """ Liefert die kleinste Box, die obj vollstaendig enthaelt. """
    if isinstance(obj, geometry.Circle):
        c = obj.getCenter()
        r = obj.getRadius()
        p1 = geometry.Point(None,c.x+r,c.y+r)
        p2 = geometry.Point(None,c.x-r,c.y-r)
        return geometry.Box(None,p1,p2)        
    else:
        y = []
        x = []
        for p in obj.getPoints():
            y.append(p.y)
            x.append(p.x)
        p1 = geometry.Point(None,max(x),max(y))
        p2 = geometry.Point(None,min(x),min(y))
        return geometry.Box(None,p1,p2)
    
class Test:
    """ Zum debuggen """
    def __init__(self):
        p1 = geometry.Point(None, -2.0,  0.0)        
        p2 = geometry.Point(None, 0.0,  2.0)
        p3 = geometry.Point(None, 2.0,  0.0)
        plg = geometry.Polygon(None,(p1,p2,p3))
        print plg
        buf = buffer(plg,1)
        bnbox = getBoundingBox(buf)
        box = geometry.Box(None, p1, p2)
        fprint([bnbox,plg,buf])

if __name__=="__main__":
	Test()


