# ****************************************************
# Name des Moduls: Geometry
# Name des Projekts: Bufferlo
#
# Autor(en) des Moduls:
#       Titus von der Malsburg
#
# Aufgaben des Moduls: siehe DocString
# Compiler fr eingebettete Dokumentation: PyDoc
#
# Datum der letzten nderung: 12.01.2003
# ****************************************************


""" Diese Bibliothek wurde uns (ein groes Dankeschn!!) von Titus
    von der Malsburg zur Verfgung gestellt, der sie fr ein Programm
    im Rahmen eines Seminars ber Sprachgenerierung entwickelt hat. Es
    sollte ein Touristeninformationssystem programmiert werden - so
    entstand ja auch die Idee zu Bufferlo.

    Das Modul enthlt Klassen, die allgemeine geometrische Objekte
    (Punkt, Linie, Kreis, Polygon) beschreiben, sowie zahlreiche Funktionen,
    die geometrische Tests auf der Grundlage dieser Klassen ausfhren:
    Abstnde messen, Winkel messen und ziehen, Schnittpunkte bestimmen.

    Fr das Bufferlo-Projekt haben wir das Modul um eine Funktion erweitert:
    createLot(). Diese Funktion gibt fr einen Punkt und eine Linie das Lot
    vom Punkt auf diese Linie zurck, sofern mglich. Auerdem haben wir
    die Funktion DPointLine() (Abstand Punkt/ Linie) berarbeitet, da uns das
    Original merkwrdige Werte zu liefern schien. Sie ist als DPointLine_Titus()
    aber noch vorhanden.
"""



#####
# Titus' kleine Geometrie Bibliothek
# begonnen 11.5.2001
#####
# translate bei Line rauswerfen?
# fuer alle objecte ein memberfield fuer points anlegen
# getPoints aendern
# pruefen ob weglassung des "letzten" punkts was ausmacht
#  * der gnuplotprinter muckt
###
import math, copy, sys
from berror import GeometryError

class Object:
    """ Hieraus erben alle geometrischen Objecte. """
    def translate(self, p):
        for pt in self.getPoints():
            pt = pt + p
    def rotate(self, w):
        """ Drehung um den Ursprung """

        w = w * math.pi / 180
        for p in self.getPoints():
            if p.x == 0 and p.y ==0:
                continue
            pt = copy.deepcopy(p)
            p.x =  pt.x*math.cos(w)+pt.y*math.sin(w)
            p.y = -pt.x*math.sin(w)+pt.y*math.cos(w)
    def getID(self):
        return self.ID

class Point(Object):
    def __init__(self, ID, x, y):
        self.x = x
        self.y = y
        self.ID = ID
    def __sub__(self, other):
        return Point(None, self.x-other.x, self.y-other.y)
    def __add__(self, other):
        return Point(None, self.x+other.x, self.y+other.y)
    def __repr__(self):
        return "(%s,%s)" % (self.x,self.y)
    def __cmp__(self, other):
        if not isinstance(other, Point):
            return 1
        if str(self.x) == str(other.x) and str(self.y) == str(other.y):
            return 0
        else:
            return 1
    def mult(self, f):
        return Point(None, self.x*f, self.y*f)
    def norm(self):
        """ Eukildische Norm """
        return DPointPoint(Point(None,0,0),self)
    def getPoints(self):
        return [self]
    def getXY(self):
        return (self.x,self.y)

class Line(Object):
    """ Die Line wird sowohl als Liniensegment zwischen den Punkten benutzt, als auch als Gerade. """
    def __init__(self, ID, p1, p2):
        self.P1 = p1
        self.P2 = p2
        self.ID = ID
    def __repr__(self):
        return `self.P1` +"-"+`self.P2`
    def __cmp__(self, other):
        if ((self.P1 == other.P1 and self.P2 == other.P2) or
            (self.P1 == other.P2 and self.P2 == other.P1)):
            return 0
        else:
            return 1
    def reverse(self):
        """ Die beiden Punkte werden vertauscht. (In manchen Zusammenhaengen ist die Richtung der Line nicht egal.) """
        pt = self.P2
        self.P2 = self.P1
        self.P1 = pt
    def length(self):
        return DPointPoint(self.P1,self.P2)
    def getXY(self):
        return [self.P1.x, self.P1.y, self.P2.x, self.P2.y]
    def getPoints(self):
        return [self.P1, self.P2]
    def getLines(self):
        return [self]
    def translate(self, p):
        self.P1 = self.P1 + p
        self.P2 = self.P2 + p
    def scaleToLength(self,l):
        """ Scaliert die Line zu einer gegebenen Laenge, wobei der ersten Punkt fix bleibt. """
        pt = self.P2 - self.P1
        if pt.norm()>1e-10:
          f = l/pt.norm()
          self.P2.x = self.P1.x + pt.x * f
          self.P2.y = self.P1.y + pt.y * f

class PolyObject(Object):
    """ Hieraus erben alle Objekte die aus mehreren Lines bestehen."""
    def __len__(self):
        return len(self.lines)
    def __repr__(self):
        str = ''
        for x in self.lines:
            str = str + `x` + ', '
        return str
    def getLines(self):
        return self.lines
    def getPoints(self):
        return self.points
    def length(self):
        """ Liefert due Summer der Laengen der Lines aus denen das Objekt besteht. """
        l=0
        for x in self.lines:
            l = l + x.length()
        return l

class Path(PolyObject):
    """ List of connected line segments """
    def __init__(self, ID, points):
        self.points = points
        self.lines = []
        self.ID = ID
        a = None
        for x in points:
            if not a:
                a = start = x
                continue
            self.lines.append(Line(None,a,x))
            a = x

class Polygon(PolyObject):
    def __init__(self, ID, points):
        self.points = points
        self.lines = []
        self.ID = ID
        a = None
        for x in points:
            if a == x:
                continue
            if not a:
                a = start = x
                continue
            self.lines.append(Line(None,a,x))
            a = x
        if not self.lines:
            raise GeometryError("Polygon.init", 4)
        if self.lines[-1].P2 != self.lines[0].P1:
            self.lines.append(Line(None, self.lines[-1].P2, self.lines[0].P1))

  
class Box(Polygon):
    """ Eigentlich ein stinknormales Polygon. Habs wegen dem Konstruktor mit den zwei Punkten gemacht. """
    def __init__(self, ID, p1, p2):
        p3 = Point(None,p1.x,p2.y)
        p4 = Point(None,p2.x,p1.y)
        Polygon.__init__(self,ID, [p1,p4,p2,p3])

    def __getitem__(self, index):
        return [self.points[0].x, self.points[0].y, self.points[2].x, self.points[2].y][index]


class Circle(Object):
    """ Besteht aus einem Mittelpunkt (center) und einem Radius. """
    def __init__(self,ID,p,r):
        self.center = p
        self.radius = r
        self.ID = ID
    def __repr__(self):
        return "((%s,%s),%s)" % (self.center.x,self.center.y,self.radius)
    def getCenter(self):
        return self.center
    def getRadius(self):
        return self.radius
    def getPoints(self):
        return [self.center]
        
def crossProdukt(p1,p2):
    return -1 * p1.x * p2.y + p1.y * p2.x

def skalarMult(sk,p):
    return Point(0, p.x*sk, p.y*sk)

def skalarProdukt(p1,p2):
    return (p1.x*p2.x + p1.y*p2.y)

def angle(ln1,ln2):
    """ Liefert den Winkel zwischen zwei Linien. """
    if ln1.P1 != ln2.P1:
        lna = Line(None,ln1.P1-Point(None,0,0),ln2.P2-(ln1.P1-ln2.P1))
    else:
        lna = copy.deepcopy(ln2)
    lnb = copy.deepcopy(ln1)
    tp = Point(None,ln1.P1.x*-1,ln1.P1.y*-1)
    lnb.translate(tp)
    lna.translate(tp)
    winkel = math.acos(round(skalarProdukt(lna.P2,lnb.P2)/(lna.P2.norm()*lnb.P2.norm()),12))
    winkel = (winkel*180/math.pi)
    return winkel

def angleClockwise(ln1,ln2):
    ln1 = copy.deepcopy(ln1)
    ln2 = copy.deepcopy(ln2)
    ln1.scaleToLength(1)
    ln2.scaleToLength(1)
    ln1.translate(Point(None,0,0)-ln1.P1)
    ln2.translate(Point(None,0,0)-ln2.P1)
    an = angle(ln1,ln2)
    ln1a = copy.deepcopy(ln1)
    ln1b = copy.deepcopy(ln1)
    ln1a.rotate(an)
    ln1b.rotate(-an)
    if angle(ln1a,ln2) < angle(ln1b,ln2):
        return an
    else:
        return 360-an

def createAngle(linie, winkel):
    lna = copy.deepcopy(linie)
    tp = Point(None,linie.P1.x*-1,linie.P1.y*-1)
    lna.translate(tp)
    lna.rotate(winkel)
    tp = tp.mult(-1)
    lna.translate(tp)
    return lna
        
#####
# geometrische Tests
##

def turns(p1,p2,p3):
    """ In welche Richtung geht die Kurve, die durch verbinden der Punkte entsteht? """
    c = round(crossProdukt(p2-p1, p3-p1),8)
    if c > 0:
        return 'left'
    elif c < 0:
        return 'right'
    else:
        return 'collinear'

def collinear(p1,p2,p3):
    """ lassen sich die drei Punkte zu einer Geraden verbinden """
    t = turns(p1,p2,p3)    
    return t == 'collinear'

def sameSide(p0, p1, line):
    """ Befinden sich zwei Punkte auf der gleichen Seite einer Geraden? Wenn einer der beiden sich auf der Gerade befindet ist's auch ok. """
    t0 = turns(line.P1,line.P2,p0)
    t1 = turns(line.P1,line.P2,p1)
    return (t0 == 'right' and t1 != 'left') or (t0 == 'left' and t1 != 'right')

def between(line, p):
    """ Liegt p zwischen den Punkten der Linie? (Damit ist nicht gemeint, ob er auf der Verbindungslinie liegt.) """
    s0,s1 = line.P1, line.P2
    if not collinear(p,s0,s1):
        return None
    else:
        #Der Punkt ist also mit der Linie colinear. Aber liegt er auch zwischen den beiden Punkten?
        if s0.x != s1.x:
            return min(s0.x,s1.x) <= p.x and max(s0.x, s1.x) >= p.x
        elif s0.y != s1.y:
            return min(s0.y,s1.y) <= p.y and max(s0.y, s1.y) >= p.y
        else:
            return s0.x == p.x and s0.y == p.y

def createLot(point, line):
    """ Berechnet das Lot vom Punkt point auf die Linie line und gibt es als Linie zurck.
        Gibt None zurck, wenn kein Lot existiert.
    """
    
    #Wir berechnen die Linien von den beiden line-Punkten zu point
    ankathete = Line(None, line.P1, point)
    gegenkathete = Line(None, line.P2, point)
    
    #Berechnung der Winkels zwischen ankathete, gegenkathete und line
    if ankathete.P1 == ankathete.P2:
        alpha = 0
    else:
        alpha = angle(ankathete, line)

    if gegenkathete.P1 == gegenkathete.P2:
        beta = 0
    else: 
        beta = angle(gegenkathete, line)
        
    #Falls beide Winkel nicht grer als 90 sind, ist ein Lot mglich
    if (round(alpha,5) <= 90) and (round(beta,5) <= 90):
        #Umrechnung in Bogenma
        alpha = alpha * math.pi / 180    

        #d ist die Lnge der Linie zwischen dem 1. Punkt der Line und dem Lotfupunkt
        d = math.cos(alpha) * ankathete.length()
        #linie_d hat den Anfangspunkt von line, aber als Endpunkt den Lotfupunkt
        linie_d = copy.deepcopy(line)
        linie_d.scaleToLength(d)                
    
        lot = Line(None,
                   Point(None, round(point.x,5),round(point.y,5)),
                   Point(None, round(linie_d.P2.x,5),round(linie_d.P2.y,5)))
    else:        
        lot= None
        
    return lot        


#####
# Abstaende von verschiedenen Objekten:
##
def DPointPoint(p1,p2):
    """ Abstand von p1 zu p2. """
    t = p1-p2
    return math.sqrt(t.x*t.x + t.y*t.y)

def DPointLine_Titus(p, l):
    """ Abstand vom Punkt p zur Line l. 
        ACHTUNG: War frher "DPointLine", gibt aber
        seltsame Werte zurck. Neue Funktion siehe unten."""
    t = crossProdukt(p-l.P1, l.P2)/ skalarProdukt(l.P2,l.P2)
    f = l.P1 + skalarMult(t, l.P2)
    return f.norm()


def DPointLine(p, l):
    """ Abstand vom Punkt p zur Linie l. """
    if collinear(p,l.P1,l.P2):
        d = 0.0
    else:
        lot = createLot(p,l)
        if lot:
            d = lot.length()
        else:
            d = min(Line(None,l.P1,p).length(),Line(None,l.P2,p).length())
    return d

    
def DLineLine(l1,l2):
    """ Abstand von der Line l1 zur Line l2. """
    d0 = DPointLine(l1.P1, l2)
    d1 = DPointLine(l1.P2, l2)
    d2 = DPointLine(l2.P1, l1)
    d3 = DPointLine(l2.P2, l1)
    return min(d0,d1,d2,d3)

def DLinePolyObject(l,pl):
    list = []
    for x in pl.getLines():
        list.append(DLineLine(x,l))
    return min(list)

def DPolyObjectPolyObject(pl1,pl2):
    list = []
    for x in pl1.getLines():
        list.append(DLinePolyObject(x,pl2))
    return min(list)

def DCirclePolyObject(c,pl):
    """ Abstand vom Circle c zur PolyLine pl. """
    d = DPointPolyObject(c.getCenter(),pl)
    d = d - c.getRadius()
    if d < 0:
        return 0
    else:
        return d

#####
# Tests auf Schnittpunkte von Objekten
##

def intPoint(ln1,ln2):
    """ Liefert den Schnittpunkt von zwei Geraden. """
    a,u = ln1.P1,ln1.P2-ln1.P1
    b,v = ln2.P1,ln2.P2-ln2.P1
    norm = u.x*v.y - u.y*v.x
    if norm==0:
        raise GeometryError("intPoint", 5)
    p = (u.mult(b.x*v.y - b.y*v.x) - v.mult(a.x*u.y - a.y*u.x)).mult(1/norm)
    return p

def intersects(obj1,obj2):
    """ Prueft, ob das erste und das zweite Objekt sich schneiden. """
    if isinstance(obj1, Circle):
        return IntObjectCircle(obj2,obj1)
    elif isinstance(obj2, Circle):
        return IntObjectCircle(obj1,obj2)
    else:
        return IntLineObjectLineObject(obj1,obj2)

def IntLineLine(line1, line2):
    s1, s2 = line1.P1, line1.P2
    t1, t2 = line2.P1, line2.P2
    if collinear(s1,s2,t1) or \
       collinear(s1,s2,t2) or \
       collinear(t1,t2,s1) or \
       collinear(t1,t2,s2):
           return between(line1,t1) or \
                  between(line1,t2) or \
                  between(line2,s1) or \
                  between(line2,s2)
    else:
        return not sameSide(s1,s2,line2) and not sameSide(t1,t2,line1)

def IntLineObjectLineObject(obj1,obj2):
    count = 0
    
    for ln1 in obj1.getLines():
        for ln2 in obj2.getLines():
            if IntLineLine(ln1,ln2):
                count += 1
                
    return count

def IntObjectCircle(obj,c):
    if DCirclePolyObject(c,obj) != 0:
        return 1
    else:
        return None

def contains(obj,ct):
    """ Es wird getestet ob das zweite Objekt das erste beinhaltet. Punkt-in-Polygon-Test nach dem Jordan curve theorem. Der Fall, wenn ein Punkt auf einer Linie des Polgons liegt kann nicht korrekt bearbeitet werden! """
    if isinstance(ct, Circle):
        return ObjectInCircle(obj,ct)
    elif isinstance(obj, Point):
        return PointInPolygon(obj,ct)
    else:
        return ObjectInPolygon(obj,ct)

def containsPartOf(obj,ct):
    """ Prueft ob das obj zumindest teilweise in ct enthalten ist. """
    if intersects(obj,ct) or contains(obj,ct):
        return 1
    else:
        return None

def PointInPolygon(point, polygon):
    i, c = 0, 0
    points = polygon.getPoints()
    j = len(points)-1
    while i < len(points):
        if (((points[i].y<=point.y) and (point.y<points[j].y)) or \
             ((points[j].y<=point.y) and (point.y<points[i].y))) and \
            (point.x < (points[j].x - points[i].x) * (point.y - points[i].y)/\
             (points[j].y - points[i].y) + points[i].x):
            c = not c
        j = i
        i = i+1
    return c

def LineInPolygon(ln,plg):
    if (PointInPolygon(ln.P1,plg) and
        PointInPolygon(ln.P2,plg) and
        not intersects(ln,plg)):
        return 1
    else:
        return None
           
def ObjectInPolygon(obj,plg):
    if isinstance(obj,Circle):
        if PointInPolygon(c.getCenter(),plg) and DCirclePolygon(c,plg) > 0:
            return 1
        else:
            return None
    else:
        for ln in obj.getLines():
            if not LineInPolygon(ln,plg):
                return None
        return 1
    
def ObjectInCircle(obj,c):
    if isinstance(obj,Circle):
        if (obj.getCenter()-c.getCenter()).norm()+obj.getRadius() > c.getRadius():
            return None
    else:
        for p in obj.getPoints():
            if (p-c.getCenter()).norm() > c.getRadius():
                return None
    return 1

#####
# Klasse die die Funktionen testet:
##
class Test:
    """ Hiermit kann man fein debuggen. """
    def __init__(self):
        p2 = Point(None, 15, 25)
        p3 = Point(None, 35, 25)
        p4 = Point(None, 15, 25)
##        p5 = Point(None, 10, 35)
##        p6 = Point(None, 10, 20)
##
##        p7 = Point(None, 5, 5)
##        p8 = Point(None, 25, 34)
##
##        p9 = Point(None, 30, 25)
##        p10 = Point(None, 45, 45)
##        
##        poly= Polygon(None,[p2, p3, p4, p5, p6])
##        ln1= Line(None, p7, p8)
        ln2 = Line(None,p2,p3)

        print "Lot= ", createLot(p4, ln2)


if __name__=="__main__":
	Test()