# ****************************************************
# Name des Moduls: PrepContainer
# Name des Projekts: Bufferlo
#
# Autor(en) des Moduls:
#       Etienne Verleih, verleih@cl.uni-heidelberg.de
#
# Aufgaben des Moduls: siehe DocString
# Compiler fr eingebettete Dokumentation: PyDoc
#
# Datum der letzten nderung: 10.01.2003
# ****************************************************


""" Die Klasse PrepContainer modeliert ein Containerobjekt fr Prpositions- und Modi-
    fikatorobjekte (siehe Modul preposition). Es enthlt zwei Listen, self.cpreps und
    self.cmods, in denen Prpositionen bzw. Modifikatoren abgelegt werden.

    Es werden diverse Methoden angeboten, um Prpositionen der Liste hinzuzufgen, zu
    lschen und zu suchen. Mit den Methoden getpreps() bzw. getmods() knnen Daten aus
    externen Quellen gelesen werden, sofern beim Initialisieren des Containers ein Loader-
    Objekt angelegt werden konnte (siehe __init__()).

    Ein Teil der Funktionalitt war fr die Verwendung in einem Prpositionseditor vor-
    gesehen, der leider nicht mehr realisiert werden konnte. Aber auch zum Laden von
    Prpositionen bzw. Modifikatoren ist ein Containerobjekt notwendig. Da beim Laden
    bestimmter Suchanfragen mehr als ein Objekt erstellt werden kann, kann ein Pr-
    positionsobjekt allein nicht die Fhigkeit bekommen, seine Daten aus externen
    Quellen zu lesen.

    Verwendete Module / Bibliotheken:    
    preposition: Fr die Definitionen von Prep- und Modobjekten
    preploader: zum Lesen von Daten aus externen Quellen
    btools.str_equals: fr das Suchen von Prpositionen im Container
    copy: fr das Erstellen einer Liste von Suchergebnissen
"""

from preposition import *
from preploader import Preploader
from btools import str_equals
import copy

class PrepContainer:
    
    def __init__(self, sourcename= "", sourcetyp= ""):
        """ Legt leere Listen fr enthaltene Prpositionen (cpreps) und Modifikatoren
            (cmods) an.
            
            Wenn sourcename und sourcetyp angegeben wurden, knnen anschlieend, sofern
            die Klasse Preploader den angegebenen sourcetyp untersttzt (siehe da), Pr-
            positionen und Modifikatoren aus einer externen Quelle names sourcename geladen
            werden. Hierzu siehe getpreps() bzw. getmods().
        """
        
        self.cpreps= []
        self.loader = None
        self.cmods= []
        
        if sourcename and sourcetyp:
            self.loader= Preploader(self, sourcename, sourcetyp)
            
    
    def add(self, prep):
        """ Wenn das bergebene Objekt vom Typ Preposition ist, wird es an die
            Prpositionsliste angehngt und Erfolg zurckgegeben. Wenn nicht, wird
            Misserfolg gemeldet und nichts weiter getan.
        """
        if isinstance(prep, Preposition):
            self.cpreps.append(prep)
            success= 1
        else:
            success= 0

        return success


    def remove(self, id= ""):
        """ Entfernt die Prposition mit der bergebenen ID. Wird keine ID angegeben,
            werden alle Prpositionen entfernt. Gibt Erfolg zurck, wenn etwas gelscht
            wurde, 0 wenn nicht.
        """
        if id:
            success= 0
            for x in self.cpreps:
                if x.id == id:
                    self.cpreps.remove(x)
                    success= 1
        else:
            self.cpreps= []
            success= 1

        return success
    

    def removeAllButOne(self, id):
        """ Entfernt alle Prpositionen bis auf die mit der ID id. Gibt Erfolg zurck,
            wenn id gefunden wurde und die Aktion ausgefhrt werden konnte, Misserfolg
            wenn nicht.
        """

        success= 0
        neu= []
        for x in self.cpreps:
            if x.id == id:
                neu.append(x)
                success= 1

        if success:
            self.cpreps= neu
                
        return success


    def searchpreps(self, name = "", id = ""):
        """ Diese Funktion durchsucht den PrepContainer nach den Prpositionen,
            auf die name ODER id zutrifft. Die gefundenen Prpositionen werden
            in einer Liste zurckgegeben (als Kopie!)
        """
        
        erg = []
        if name or id:
            for prep in self.cpreps:
                if str_equals(name, prep.name):
                    erg.append(copy.deepcopy(prep))
                elif str_equals(id, prep.id):
                    erg.append(copy.deepcopy(prep))
                    #Da eine ID einmalig ist, sind wir ein bichen brutal...
                    break

        return erg        

                    
    def getpreps(self, name= "", id= "", projection= ""):
        """ Liest die Prpositionen, auf die id oder name und jeweils projection
            ("Aufriss" oder "Grundriss") zutrifft, aus einer externen Quelle, benutzt dazu
            das Preploader-Objekt (zur Definition einer externen Quelle siehe __init__()).
            Wenn kein Parameter angegeben wird, werden alle Prpositionen geladen.
            
            Es wird 1 zurckgegeben, wenn erfolgreich Daten gelesen wurden, 0 bei Misserfolg.
            War das Laden erfolgreich, so stehen die gelesenen Prps anschlieend in self.cpreps
        """
            
        success= 0

        if self.loader:
            success= self.loader.getpreps(name, id, projection)
        return success            
     

    def getmods(self, name= "", id= "", projection= ""):
        """ Liest die Modifikatoren, auf die id oder name und jeweils projection ("Aufriss"
            oder "Grundriss")zutrifft, aus einer externen Quelle, benutzt dazu das Preploader-
            Objekt (zur Definition einer externen Quelle siehe __init__()). Wenn kein Parameter
            angegeben wird, werden alle Modifikatoren geladen.
            
            Es wird 1 zurckgegeben, wenn erfolgreich Daten gelesen wurden, 0 bei Misserfolg.
            War das Laden erfolgreich, so stehen die gelesenen Mods anschlieend in self.cmods
        """
            
        success= 0

        if self.loader:
            success= self.loader.getmods(name, id, projection)
        return success
    

    def __repr__(self):
        """ bergibt alle Prpositionen im Container als String, der dann ausgegeben werden kann """

        ausgabe= ""        
        for x in self.cpreps:
            ausgabe += x.__repr__()
            ausgabe += "\n\n------\n\n"
            
        return ausgabe
