# ****************************************************
# Name des Moduls: XMLParser
# Name des Projekts: Bufferlo
#
# Autor(en) des Moduls:
#       Etienne Verleih, verleih@cl.uni-heidelberg.de
#
# Aufgaben des Moduls: siehe DocString
# Compiler fr eingebettete Dokumentation: PyDoc
#
# Datum der letzten nderung: 09.01.2003
# ****************************************************


""" Diese Bibliothek soll die Verwendung des SAX-XML-Parsers von Python vereinfachen.
    Sie enthlt eine Klasse XMLParser, die sich um die hlichen Details kmmert.

    Zustzlich gibt es zwei in diesem Zusammenhang ntzliche Funktionen:

    - normalize_whitespace(), die ihrem Namen alle Ehre macht und berflssige
    Leerzeichen aus einem String entfernt. Auerdem fhrt sie eine Umwandlung von
    Unicode in ASCII durch. Sehr ntzlich, umdie Rckgaben des Parsers
    zurechtzustutzen.

    - getAttr(), fr die Attribute von XML-Tags. Liefert fr einen bergebenen Namen
    des Wert des Attributs zurck und wandelt ihn vorher ebenfalls von Unicode in
    ASCII um.
   
    Wahrscheinlich erfordert der Sax-Parser eine Python-Version >= 2.2 !!!
"""

from xml.sax import make_parser
from xml.sax.handler import feature_namespaces
import string


def normalize_str(text):
    """ Diese Funktion entfernt alle berflssigen Leerzeichen aus
        einem bergebenen String und wandelt Unicode in ASCII um.
        Sie bentigt die String-Bibliothek, die daher importiert wird.
    """
        
    text= string.join(string.split(text), ' ')
    text= text.encode('Latin-1')
    return text


def getAttr(attrs, name):
    """ Diese Funktion gibt aus den Attributen eines XML-Tags (in attrs)
        den Wert des Attributs name zurck. Dieser wird vorher von Unicode
        in ASCII umgewandelt.
    """
    
    wert= attrs.getValue(name)
    wert= wert.encode('Latin-1')
    return wert


class XMLParser:
    """ Diese Klasse erleichtert das Verwenden des SAX-XML-Parsers. 
    
        Verwendung:
        Zum Erstellen eines neuen Parserobjekts fr eine bestimmte Datei: __init__()
        Zum Parsen: parse()
        
        Bentigte Bibliotheken:
        Von xml.sax wird make_parser zum Anlegen eines Parserobjekts bentigt.
        Von xml.sax.handler wird feature_namespaces bentigt, um dem Parser mitzuteilen, dass
        keine XML-Namespaces gebraucht werden.
    """

    def __init__(self, handler, filename):
        """ Zum Erstellen eines Parserobjekts werden zwei Angaben bentigt:
            
              der Name der Datei, die geparst werden soll: filename
            
              ein Handlerobjekt, dass die vom Parser gelieferten Daten verarbeitet: handler
              (dieses Objekt muss von xml.sax.handler.ContentHandler abgeleitet sein und
              beinhaltet Methoden, die beim Eintreten bestimmter Ereignisse beim Parsen vom
              Parser automatisch aufgerufen werden - siehe die Python-Dokumentation.)
        """
        
        self.handler= handler
        self.filename= filename


    def parse(self):        
        """ Startet den Parsevorgang fr die bei init() angegebene Datei. Zum Verarbeiten
            der gelesenen Daten wird der gleichfalls angegebene Handler verwendet. """
        
        #Einen Parser erstellen
        parser = make_parser()

        #Diesem sagen, dass wir nicht an XML-Namespaces interessiert sind
        parser.setFeature(feature_namespaces, 0)

        #Ihn mit unserem Handler bekannt machen
        parser.setContentHandler(self.handler)

        #Und: Parsen!
        parser.parse(self.filename)
        